
# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""

from typing import Any
from typing import Dict
import json
import tornado

from neuro_san.service.http.logging.http_logger import HttpLogger
from neuro_san.internals.interfaces.startable import Startable
from neuro_san.service.utils.service_resources import ServiceResources


class ResourcesUsageLogger(Startable):
    """
    Class for periodic logging of server run-time resource usage:
    file descriptors and open inet connections on server port.
    """

    def __init__(self, log_interval_seconds: int, http_port: int, logger: HttpLogger):
        """
        Constructor
        :param log_interval_seconds: interval in seconds between logging resource usage
        :param http_port: http port to use
        :param logger: HttpLogger instance for logging
        """
        self.log_interval_seconds = log_interval_seconds
        self.http_port: int = http_port
        self.logger: HttpLogger = logger
        self.periodic_callback = tornado.ioloop.PeriodicCallback(
            self.run_resources_usage,
            self.log_interval_seconds * 1000
        )

    def log_resources_usage(self):
        """
        Log current usage of server run-time resources:
        file descriptors and open inet connections on server port.
        """
        # Get used file descriptors:
        fd_dict, soft_limit, hard_limit = ServiceResources.get_fd_usage()
        sock_classes = ServiceResources.classify_sockets(self.http_port)
        log_dict: Dict[str, Any] = {
            "soft_limit": soft_limit,
            "hard_limit": hard_limit,
            "file_descriptors": fd_dict,
            "sockets": sock_classes
        }
        self.logger.info({}, "Used: %s", json.dumps(log_dict, indent=4))

    async def run_resources_usage(self):
        """
        Execute collecting and logging of server run-time resources
        in on-blocking mode w.r.t. server event loop.
        This is done because enumerating of some system resources
        could be relatively slow.
        """
        loop = tornado.ioloop.IOLoop.current()
        return await loop.run_in_executor(None, self.log_resources_usage)

    def start(self):
        """
        Start periodic logging of resource usage.
        """
        self.periodic_callback.start()




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

# DEF - tangle to be resolved later
from neuro_san.service.generic.async_agent_service_provider import AsyncAgentServiceProvider


class AgentAuthorizer:
    """
    Interface for authorizing agent specifics given metadata from a request.
    """

    async def allow_agent(self, agent_name: str, metadata: Dict[str, Any]) -> Tuple[bool, AsyncAgentServiceProvider]:
        """
        Is the request allowed for this agent?

        :param agent_name: name of an agent
        :param metadata: metadata from the request
        :return: a tuple of:
                * True if metadata says user is authrorized to route requests is allowed for this agent
                  False otherwise
                * instance of AsyncAgentService if it exists.  None otherwise
        """
        raise NotImplementedError

    async def list_agents(self, metadata: Dict[str, Any]) -> List[str]:
        """
        What is the list of allowed agents for this request?
        :param metadata: metadata from the request
        :return: a list of agent names allowed for this request
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT


class AgentNameMapper:
    """
    An abstract policy defining conversion
    between agent name as specified in a manifest file
    and a file path (relative to registry root directory) to this agent definition file.
    """

    def agent_name_to_filepath(self, agent_name: str) -> str:
        """
        Converts an agent name from manifest file to file path to this agent definition file.
        """
        raise NotImplementedError()

    def filepath_to_agent_network_name(self, filepath: str) -> str:
        """
        Converts a file path to agent definition file (relative to registry root directory)
        to agent network name identifying it to the service.
        """
        raise NotImplementedError()




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from neuro_san.internals.interfaces.startable import Startable


class StorageUpdater(Startable):
    """
    Interface for specific updating jobs that the Watcher performs.
    """

    def start(self):
        """
        Perform start up.
        """
        raise NotImplementedError

    def get_update_period_in_seconds(self) -> int:
        """
        :return: An int describing how long this updater ideally wants to go between
                calls to update_storage().
        """
        raise NotImplementedError

    def needs_updating(self, time_now_in_seconds: float) -> bool:
        """
        :param time_now_in_seconds: The current time in seconds.
                    We expect this to be from time.time()
        :return: True if this instance needs updating. False otherwise.
        """
        raise NotImplementedError

    def update_storage(self):
        """
        Perform an update
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict

from neuro_san.service.mcp.mcp_errors import McpError
from neuro_san.service.mcp.util.mcp_request_util import McpRequestUtil


class McpErrorsUtil:
    """
    Utility class for generating MCP protocol and tool error responses.
    """

    @classmethod
    def get_protocol_error(cls, request_id, error: McpError, extra_msg: str = None) -> Dict[str, Any]:
        """
        Generate a standard MCP protocol error response.
        :param request_id: MCP request id;
        :param error: MCPError enum value;
        :param extra_msg: Optional extra message to append to the standard error message;
        :return: json dictionary with error in MCP format suitable for sending to the client.
        """
        msg: str = error.str_label
        if extra_msg is not None:
            msg = f"{msg}: {extra_msg}"
        return {
            "jsonrpc": "2.0",
            # Appease code scanning tools by escaping the id field:
            "id": McpRequestUtil.safe_request_id(request_id),
            "error": {
                "code": error.num_value,
                "message": McpRequestUtil.safe_message(msg)
            }
        }

    @classmethod
    def get_tool_error(cls, request_id, error_msg: str) -> Dict[str, Any]:
        """
        Generate a standard MCP tool error response.
        :param request_id: MCP request id;
        :param error_msg: Error message to send to the client;
        :return: json dictionary with tool error in MCP format suitable for sending to the client.
        """
        return {
            "jsonrpc": "2.0",
            # Appease code scanning tools by escaping the id field:
            "id": McpRequestUtil.safe_request_id(request_id),
            "result": {
                "content": [
                    {
                        "type": "text",
                        "text": McpRequestUtil.safe_message(error_msg)
                    }
                ],
                "isError": True
            }
        }

    @classmethod
    def is_error(cls, response_dict: Dict[str, Any]) -> bool:
        """
        Check if the given MCP response dictionary represents an error.
        :param response_dict: MCP response dictionary;
        :return: True if the response is an error, False otherwise.
        """
        # Check for protocol error first:
        protocol_error = response_dict["error"]
        if protocol_error is not None and len(protocol_error) > 0:
            return True
        # Check for tool error:
        has_error: bool = response_dict.get("result", {}).get("isError", False)
        return has_error




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from collections.abc import Generator
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any
from typing import Dict
from typing import Optional

from langchain_core.tracers.context import register_configure_hook

from neuro_san.internals.run_context.langchain.token_counting.llm_token_callback_handler \
    import LlmTokenCallbackHandler


llm_token_callback_var: ContextVar[Optional[LlmTokenCallbackHandler]] = (
        ContextVar("llm_token_callback", default=None)
    )
register_configure_hook(llm_token_callback_var, inheritable=True)


@contextmanager
def get_llm_token_callback(llm_infos: Dict[str, Any]) -> Generator[LlmTokenCallbackHandler, None, None]:
    """Get llm token callback.

    Get context manager for tracking usage metadata across chat model calls using
    "AIMessage.usage_metadata".

    This class is a modification of LangChain’s "UsageMetadataCallbackHandler":
    - https://python.langchain.com/api_reference/_modules/langchain_core/callbacks/usage.html
    #get_usage_metadata_callback

    :param llm_infos: Dictionary containing configuration or metadata about the LLM
                      (e.g., model name, class (provider), token cost).
    :return: A generator-based context manager that yields an `LlmTokenCallbackHandler`
             for tracking token usage within the context.
    """
    # Create a new callback handler instance for tracking token usage
    cb = LlmTokenCallbackHandler(llm_infos)

    # Set the context variable to the newly created callback handler
    llm_token_callback_var.set(cb)

    # Yield the callback handler to the context block
    yield cb

    # Reset the context variable to None when the context exits
    llm_token_callback_var.set(None)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider


class AgentStorageSource:
    """
    Interface onto AgentNetworkStorage so that there is not a tangle with the service side.
    """

    def get_agent_network_provider(self, agent_name: str) -> AgentNetworkProvider:
        """
        Get AgentNetworkProvider for a specific agent
        :param agent_name: name of an agent
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

from logging import getLogger
from logging import Logger

from neuro_san.internals.validation.network.abstract_network_validator import AbstractNetworkValidator


class ToolboxNetworkValidator(AbstractNetworkValidator):
    """
    AbstractNetworkValidator for toolbox references.
    """

    def __init__(self, tools: Dict[str, Any]):
        """
        Constructor

        :param tools: A dictionary of tools, as read in from a toolbox_info.hocon file
        """
        self.logger: Logger = getLogger(self.__class__.__name__)
        self.tools: Dict[str, Any] = tools

    def validate_name_to_spec_dict(self, name_to_spec: Dict[str, Any]) -> List[str]:
        """
        Validate the agent network, specifically in the form of a name -> agent spec dictionary.

        :param name_to_spec: The name -> agent spec dictionary to validate
        :return: A list of error messages
        """
        errors: List[str] = []

        self.logger.info("Validating toolbox agents...")

        for agent_name, agent in name_to_spec.items():
            if agent.get("instructions") is None:  # This is a toolbox agent
                if self.tools is None or not isinstance(self.tools, Dict):
                    errors.append(f"Toolbox is unavailable. Cannot create Toolbox agent '{agent_name}'.")
                elif agent_name not in self.tools:
                    errors.append(f"Toolbox agent '{agent_name}' has no matching tool in toolbox.")
                elif agent.get("tools"):
                    errors.append(
                        "Toolbox agent cannot have 'tools'. "
                        f"[{agent.get('tools')}] cannot be under Toolbox agent '{agent_name}'"
                    )

        return errors




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict
from typing import Tuple

from neuro_san.service.http.logging.http_logger import HttpLogger
from neuro_san.service.utils.mcp_server_context import McpServerContext
from neuro_san.service.mcp.interfaces.client_session import ClientSession
from neuro_san.service.mcp.util.mcp_request_util import McpRequestUtil


class McpInitializeProcessor:
    """
    Class implementing client session initialization.
    """
    def __init__(self, mcp_context: McpServerContext, logger: HttpLogger):
        self.logger: HttpLogger = logger
        self.mcp_context: McpServerContext = mcp_context

    async def initialize_handshake(
            self,
            request_id,
            metadata: Dict[str, Any],
            params: Dict[str, Any]) -> Tuple[Dict[str, Any], str]:
        """
        Process initial protocol handshake.
        :param request_id: MCP request id;
        :param metadata: http-level request metadata;
        :param params: dictionary with handshake parameters;
        :return: json dictionary with handshake response
        """
        # Currently, we do not use any parameters from the client
        # for protocol version or capabilities negotiation.
        # We simply return the server capabilities.
        # Also: we don't look at possible session ID present in the incoming request.
        # Future versions may implement more complex negotiation logic.

        _ = params
        # Create new client session:
        session: ClientSession = self.mcp_context.get_session_policy().create_session()
        session_id: str = None
        if session:
            session_id = session.get_id()
            self.logger.info(metadata, "Created new MCP client session with id: %s", session_id)

        response_dict: Dict[str, Any] = McpRequestUtil.get_handshake_response(request_id)
        return response_dict, session_id

    async def activate_session(
            self,
            session_id: str,
            metadata: Dict[str, Any]) -> Dict[str, Any]:
        """
        Activate existing client session.
        :param session_id: session id to activate;
        :
        :param metadata: http-level request metadata;
        :return: True if successful;
                 False if session with given id does not exist
        """
        success: bool = self.mcp_context.get_session_policy().activate_session(session_id)
        if not session_id:
            session_id = "N/A"
        if success:
            self.logger.info(metadata,
                             "Activated MCP client session with id: %s",
                             session_id)
        else:
            self.logger.info(metadata,
                             "Failed to activate MCP client session with id: %s - session not found",
                             session_id)
        return success




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

import logging
import os

from leaf_server_common.logging.logging_setup import setup_logging
from leaf_server_common.server.grpc_metadata_forwarder import GrpcMetadataForwarder


class AgentServerLogging:
    """
    Common logging setup for the agent server threads.
    """

    def __init__(self, server_name_for_logs: str,
                 forwarded_request_metadata_str: str,
                 logging_config: Dict[str, Any] = None):
        """
        Constructor

        :param server_name_for_logs: The server name (source) to be used in structured logging messages
        :param forwarded_request_metadata_str: A space-delimited string of http header metadata
                            whose key/value pairs are to be forwarded on to the logging system.
                            Note that individual keys must be snake_case. No capitals.
                            (I guess per HTTP rules).
        :param logging_config: A dictionary of configuration parameters for the logging system
        """
        self.server_name_for_logs: str = server_name_for_logs
        self.forwarded_request_metadata: List[str] = forwarded_request_metadata_str.split(" ")
        self.logging_config: Dict[str, Any] = logging_config

    def get_forwarder(self) -> GrpcMetadataForwarder:
        """
        :return: A GrpcMetadataForwarder instance initialized with
                 the list of forwarded request metadata keys
        """
        return GrpcMetadataForwarder(self.forwarded_request_metadata)

    def setup_logging(self, metadata: Dict[str, str] = None, request_id: str = "None"):
        """
        Set up logging for agent server threads.

        :param metadata: An optional dictionary with actual metadata
        :param request_id: An optional request_id string.  Default is "None".
        """

        # Need to initialize the forwarded metadata default values before our first
        # call to a logger (which is below!).
        extra_logging_defaults: Dict[str, str] = {
            "source": self.server_name_for_logs,
            "user_id": "None",
            "request_id": request_id,
        }
        if len(self.forwarded_request_metadata) > 0:
            for key in self.forwarded_request_metadata:
                if metadata is not None:
                    extra_logging_defaults[key] = metadata.get(key, "None")
                else:
                    extra_logging_defaults[key] = "None"

        current_dir: str = os.path.dirname(os.path.abspath(__file__))
        setup_logging(self.server_name_for_logs,
                      default_log_dir=current_dir,
                      log_level_env="AGENT_SERVICE_LOG_LEVEL",
                      extra_logging_fields_defaults=extra_logging_defaults,
                      logging_config=self.logging_config)

        # This module within openai library can be quite chatty w/rt http requests
        logging.getLogger("httpx").setLevel(logging.WARNING)

    def redact_per_env_var(self, input_str: str, env_var_name: str) -> str:
        """
        Redact a string using the value of an environment variable.

        :param input_str: The string to redact
        :param env_var_name: The name of the environment variable
        :return: The redacted string
        """
        value: str = f"'{input_str}'"

        # Determine the value of the environment variable
        slice_str: str = os.environ.get(env_var_name)
        if slice_str is not None and len(slice_str) > 0:

            # Deterimine the extent and direction of the slice
            try:
                marker_slice: int = int(slice_str)

                if marker_slice == 0:
                    value = "<redacted>"
                elif marker_slice < 0:
                    one_slice: str = input_str[marker_slice:]
                    prefix: str = ""
                    if -marker_slice < len(input_str):
                        prefix = "..."
                    value = f"'{prefix}{one_slice}'"
                else:
                    one_slice: str = input_str[:marker_slice]
                    suffix: str = ""
                    if marker_slice < len(input_str):
                        suffix = "..."
                    value = f"'{one_slice}{suffix}'"
            except ValueError:
                # Not an int
                logging.getLogger(self.__class__.__name__).info(
                    "Value for %s: '%s' needs to be integer or empty string.",
                    env_var_name, slice_str)

        return value




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
# Needed for method referencing different instance of the same class
# See https://stackoverflow.com/questions/33533148/how-do-i-type-hint-a-method-with-the-type-of-the-enclosing-class
from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List

from langchain_core.messages.base import BaseMessage

from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.run_context.interfaces.agent_spec_provider import AgentSpecProvider
from neuro_san.internals.run_context.interfaces.run import Run


class RunContext(AgentSpecProvider):
    """
    Interface supporting high-level LLM usage.
    """

    async def create_resources(self, agent_name: str,
                               instructions: str,
                               assignments: str,
                               tool_names: List[str] = None):
        """
        Creates resources to be used during a run of an agent.
        The result is stored as a member in this instance for future use.
        :param agent_name: String name for the agent.
        :param instructions: string instructions that are used to create the agent
        :param assignments: string assignments of function parameters that are used as input
        :param tool_names: The list of registered tool names to use.
                    Default is None implying no tool is to be called.
        """
        raise NotImplementedError

    async def submit_message(self, user_message: str) -> Run:
        """
        Submits a message to create a run.
        :param user_message: The message to submit
        :return: The Run instance which is processing the agent's message
        """
        raise NotImplementedError

    async def wait_on_run(self, run: Run, journal: Journal = None) -> Run:
        """
        Loops on the given run's status for service-side processing
        to be done.
        :param run: The Run instance to wait on
        :param journal: The Journal which captures the "thinking" messages.
        :return: An potentially updated Run instance
        """
        raise NotImplementedError

    async def get_response(self) -> List[BaseMessage]:
        """
        :return: The list of messages from the instance's thread.
        """
        raise NotImplementedError

    async def submit_tool_outputs(self, run: Run, tool_outputs: List[Dict[str, Any]]) -> Run:
        """
        :param run: The Run instance handling the execution of the agent
        :param tool_outputs: The tool outputs to submit
                The component dictionaries can have the following keys:
                    "origin"        A List of origin dictionaries indicating the origin of the run.
                    "output"        A string representing the output of the tool call
                    "sly_data"      Optional sly_data dictionary that might have returned from an external tool.
                    "tool_call_id"  The string id of the tool_call being executed
        :return: A potentially updated run instance handle
        """
        raise NotImplementedError

    async def delete_resources(self, parent_run_context: RunContext = None):
        """
        Cleans up the service-side resources associated with this instance
        :param parent_run_context: A parent RunContext perhaps the same instance,
                        but perhaps not.  Default is None
        """
        raise NotImplementedError

    def get_agent_tool_spec(self) -> Dict[str, Any]:
        """
        :return: the dictionary describing the data-driven agent
        """
        raise NotImplementedError

    def get_invocation_context(self) -> InvocationContext:
        """
        :return: The InvocationContext policy container that pertains to the invocation
                    of the agent.
        """
        raise NotImplementedError

    def get_chat_context(self) -> Dict[str, Any]:
        """
        :return: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
                Can be None when a new conversation has been started.
        """
        raise NotImplementedError

    def get_origin(self) -> List[Dict[str, Any]]:
        """
        :return: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        raise NotImplementedError

    def update_invocation_context(self, invocation_context: InvocationContext):
        """
        Update internal state based on the InvocationContext instance passed in.
        :param invocation_context: The context policy container that pertains to the invocation
        """
        raise NotImplementedError

    def get_journal(self) -> Journal:
        """
        :return: The Journal associated with the instance
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""


class ClientSession:
    """
    Interface representing a client session with the MCP service.
    """

    def get_id(self) -> str:
        """
        Get the session id.
        """
        raise NotImplementedError

    def is_active(self) -> bool:
        """
        Check if the session is active.
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.graph.interfaces.agent_tool_factory import AgentToolFactory
from neuro_san.internals.graph.interfaces.callable_activation import CallableActivation
from neuro_san.internals.graph.registry.activation_factory import ActivationFactory
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.front_man import FrontMan
from neuro_san.internals.run_context.interfaces.agent_network_inspector import AgentNetworkInspector
from neuro_san.internals.run_context.interfaces.run_context import RunContext


class AgentToolRegistry(AgentNetworkInspector, AgentToolFactory):
    """
    Puts together an AgentNetwork data-only spec with an ActivationFactory
    so that a single entity can handle both interfaces.
    """

    def __init__(self, agent_network: AgentNetwork):
        """
        Constructor

        :param agent_network: The AgentNetwork configuration to base all our info on
        """
        self.agent_network: AgentNetwork = agent_network
        self.factory = ActivationFactory(self.agent_network)

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def create_agent_activation(self, parent_run_context: RunContext,
                                parent_agent_spec: Dict[str, Any],
                                name: str,
                                sly_data: Dict[str, Any],
                                arguments: Dict[str, Any] = None,
                                factory: AgentToolFactory = None,
                                invocation: str = None) -> CallableActivation:
        """
        Create an active node for an agent from its spec.
        This is how CallableActivations create other CallableActivations.

        :param parent_run_context: The RunContext of the agent calling this method
        :param parent_agent_spec: The spec of the agent calling this method.
        :param name: The name of the agent to get out of the registry
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be an empty dictionary.
        :param arguments: A dictionary of arguments for the newly constructed agent
        :param factory: A factory that will be used to create the agent tool
        :param invocation: The invocation style of the activation.
        :return: The CallableActivation agent referred to by the name.
        """
        return self.factory.create_agent_activation(parent_run_context, parent_agent_spec,
                                                    name, sly_data, arguments, self, invocation)

    def create_front_man(self,
                         sly_data: Dict[str, Any] = None,
                         parent_run_context: RunContext = None) -> FrontMan:
        """
        Find and create the FrontMan for DataDrivenChat.

        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be an empty dictionary.
        :param parent_run_context: A RunContext instance
        """
        return self.factory.create_front_man(sly_data, parent_run_context, self)

    def get_agent_tool_path(self) -> str:
        """
        :return: The path under which tools for this registry should be looked for.
        """
        return self.factory.get_agent_tool_path()

    def get_config(self) -> Dict[str, Any]:
        """
        :return: The entire config dictionary given to the instance.
        """
        return self.agent_network.get_config()

    def get_agent_tool_spec(self, name: str) -> Dict[str, Any]:
        """
        :param name: The name of the agent tool to get out of the registry
        :return: The dictionary representing the spec registered agent
        """
        return self.agent_network.get_agent_tool_spec(name)

    def get_name_from_spec(self, agent_spec: Dict[str, Any]) -> str:
        """
        :param agent_spec: A single agent to register
        :return: The agent name as per the spec
        """
        return self.agent_network.get_name_from_spec(agent_spec)

    def find_front_man(self) -> str:
        """
        :return: A single tool name to use as the root of the chat agent.
                 This guy will be user facing.  If there are none or > 1,
                 an exception will be raised.
        """
        return self.agent_network.find_front_man()

    def get_agent_llm_info_file(self) -> str:
        """
        :return: The absolute path of agent llm info file for llm extension.
        """
        return self.agent_network.get_agent_llm_info_file()




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from typing import Union

from logging import Logger
from logging import getLogger

from langchain_core.tools.base import BaseTool

from neuro_san.internals.interfaces.async_agent_session_factory import AsyncAgentSessionFactory
from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.agent_message import AgentMessage
from neuro_san.internals.run_context.interfaces.agent_network_inspector import AgentNetworkInspector
from neuro_san.internals.run_context.interfaces.tool_caller import ToolCaller
from neuro_san.internals.run_context.langchain.core.langchain_openai_function_tool \
    import LangChainOpenAIFunctionTool
from neuro_san.internals.run_context.langchain.mcp.langchain_mcp_adapter import LangChainMcpAdapter
from neuro_san.internals.run_context.utils.external_agent_parsing import ExternalAgentParsing
from neuro_san.internals.run_context.utils.external_tool_adapter import ExternalToolAdapter


class BaseToolFactory:
    """
    Creates langchain BaseTools.
    """

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def __init__(self,
                 tool_caller: ToolCaller,
                 invocation_context: InvocationContext,
                 journal: Journal):
        """
        Constructor

        :param tool_caller: The ToolCaller creating tools
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param journal: The journal to use when sending framework-level messages to the client
        """
        self.tool_caller: ToolCaller = tool_caller
        self.invocation_context: InvocationContext = invocation_context
        self.journal: Journal = journal
        self.logger: Logger = getLogger(self.__class__.__name__)

    async def create_base_tool(self, name: str) -> Union[BaseTool, List[BaseTool]]:
        """
        Create base tools for the agent to call.
        :param name: The name of the tool to create
        :return: The BaseTools associated with the name
        """

        # Check our own local inspector. Most tools live in the neighborhood.
        inspector: AgentNetworkInspector = self.tool_caller.get_inspector()
        agent_spec: Dict[str, Any] = inspector.get_agent_tool_spec(name)

        if agent_spec is None:
            return await self.create_external_tool(name)

        return await self.create_internal_tool(name, agent_spec)

    async def create_external_tool(self, name: Union[str, Dict[str, Any]]) -> Union[BaseTool, List[BaseTool]]:
        """
        Create external agent/tool.
        :param name: The name of the external agent/tool
        :return: External agent as base tools
        """

        if not isinstance(name, (dict, str)):
            raise TypeError(f"Tools must be string or dict, got {type(name)}")

        # Handle MCP-based tool as external tool
        if ExternalAgentParsing.is_mcp_tool(name):
            return await self.create_mcp_tool(name)

        # See if the agent name given could reference an external agent.
        if not ExternalAgentParsing.is_external_agent(name):
            return None

        # Use the ExternalToolAdapter to get the function specification
        # from the service call to the external agent.
        # We should be able to use the same BaseTool for langchain integration
        # purposes as we do for any other tool, though.
        # Optimization:
        #   It's possible we might want to cache these results somehow to minimize
        #   network calls.
        session_factory: AsyncAgentSessionFactory = self.invocation_context.get_async_session_factory()
        adapter = ExternalToolAdapter(session_factory, name)
        try:
            function_json: Dict[str, Any] = await adapter.get_function_json(self.invocation_context)
            return self.create_function_tool(function_json, name)
        except ValueError as exception:
            # Could not reach the server for the external agent, so tell about it
            message: str = f"Agent/tool {name} was unreachable. Not including it as a tool.\n"
            message += str(exception)
            agent_message = AgentMessage(content=message)
            await self.journal.write_message(agent_message)
            self.logger.info(message)
            return None

    async def create_internal_tool(self, name: str, agent_spec: Dict[str, Any]) -> BaseTool:
        """
        Create internal agent/tool.
        :param name: The name of the agent or coded tool
        :return: Agent as base tools
        """

        toolbox: str = agent_spec.get("toolbox")

        # Handle toolbox-based tools
        if toolbox:
            return await self.create_toolbox_tool(toolbox, agent_spec, name)

        # Handle coded tools
        function_json: Dict[str, Any] = agent_spec.get("function")
        if function_json is None:
            return None

        return self.create_function_tool(function_json, name)

    async def create_mcp_tool(self, mcp_info: Union[str, Dict[str, Any]]) -> List[BaseTool]:
        """
        Create MCP tools from the provided MCP configuration.

        The configuration can be one of:
        - **String**: A URL to an MCP server.
        Valid values start with "https://mcp" or end with "/mcp" or "/mcp/".
        - **Dictionary**:
            - "server" (str): MCP server URL.
            - "tools" (List[str], optional): List of tool names to allow from the server.

        :param mcp_info: MCP server URL (string) or a configuration dictionary
        :return: A list of MCP tools as base tools
        """
        # By default, assume no allowed tools. This may get updated below or in the LangChainMcpAdadter.
        allowed_tools: List[str] = None
        # Get HTTP headers from sly_data if available
        http_headers: Dict[str, Any] = self.tool_caller.get_sly_data().get("http_headers", {})

        if isinstance(mcp_info, str):
            server_url: str = mcp_info
        else:
            server_url = mcp_info.get("url")
            allowed_tools = mcp_info.get("tools")

        # Get specific headers for the MCP server if available
        headers: Dict[str, Any] = http_headers.get(server_url)

        try:
            mcp_adapter = LangChainMcpAdapter()
            mcp_tools: List[BaseTool] = await mcp_adapter.get_mcp_tools(server_url, allowed_tools, headers)

        # MCP errors are nested exceptions.
        except ExceptionGroup as nested_exception:
            # Could not reach the MCP server
            message: str = f"The URL {server_url} was unreachable. Not including it as a tool.\n"
            message += self.get_exception_details(nested_exception)
            agent_message = AgentMessage(content=message)
            await self.journal.write_message(agent_message)
            self.logger.info(message)
            return None

        # The allowed tools list might have been updated by the MCP adapter
        allowed_tools: List[str] = mcp_adapter.client_allowed_tools
        tool_names: List[str] = [tool.name for tool in mcp_tools]
        invalid_names: Set[str] = set(allowed_tools) - set(tool_names)
        # Check if there are invalid tool names in the list.
        if invalid_names:
            message = f"The following tools cannot be found in {server_url}: {invalid_names}"
            agent_message = AgentMessage(content=message)
            await self.journal.write_message(agent_message)
            self.logger.info(message)

        return mcp_tools

    def get_exception_details(self, exception, indent=0) -> str:
        """
        Recursively extract detailed information from nested exceptions.

        This function handles both regular exceptions and ExceptionGroup instances
        (introduced in Python 3.11) which can contain multiple nested exceptions.
        It creates a human-readable, hierarchical representation of all exceptions
        in the error chain.

        :param exception: The exception to analyze. Can be any Exception type,
                            including ExceptionGroup instances that contain multiple
                            nested exceptions.
        :parm indent: The current indentation level for formatting.
                                Each recursive call increases this by 1 to create
                                a visual hierarchy. Defaults to 0.

        :return: A formatted string containing the exception type, message, and
                any nested sub-exceptions with proper indentation to show the
                hierarchy. Each line ends with a newline character.

        Note:
            This function is particularly useful for debugging MCP (Model Context Protocol)
            errors and other complex exception scenarios where multiple errors can occur
            simultaneously and get wrapped in ExceptionGroup containers.
        """

        # Create indentation string based on current nesting level
        # Each level adds 2 spaces for visual hierarchy
        spaces: str = "  " * indent

        # Start building the message with exception type and description
        # Format: "ExceptionType: exception message"
        message: str = f"{spaces}{type(exception).__name__}: {exception}\n"

        # Check if this exception is an ExceptionGroup (Python 3.11+ feature)
        # ExceptionGroup can contain multiple exceptions that occurred simultaneously
        if isinstance(exception, ExceptionGroup):
            # Iterate through each sub-exception in the group
            for i, sub_exc in enumerate(exception.exceptions):
                # Add a header for each sub-exception with 1-based numbering
                message += f"{spaces}Sub-exception {i+1}:\n"

                # Recursively process the sub-exception with increased indentation
                # This handles cases where sub-exceptions might themselves be ExceptionGroups
                message += self.get_exception_details(sub_exc, indent + 1)

        return message

    async def create_toolbox_tool(self, toolbox: str, agent_spec: Dict[str, Any], name: str) -> BaseTool:
        """Create tool from toolbox"""

        toolbox_factory: ContextTypeToolboxFactory = self.invocation_context.get_toolbox_factory()
        try:
            tool_from_toolbox = toolbox_factory.create_tool_from_toolbox(toolbox, agent_spec.get("args"), name)
            # If the tool from toolbox is base tool or list of base tool, return the tool as is
            # since tool's definition and args schema are predefined in these the class of the tool.
            if isinstance(tool_from_toolbox, BaseTool) or (
                isinstance(tool_from_toolbox, list) and
                all(isinstance(tool, BaseTool) for tool in tool_from_toolbox)
            ):
                return tool_from_toolbox

            # Otherwise, it is a shared coded tool.
            return self.create_function_tool(tool_from_toolbox, name)

        except ValueError as tool_creation_exception:
            # There are errors in tool creation process
            message: str = f"Failed to create Agent/tool '{name}': {tool_creation_exception}"
            agent_message = AgentMessage(content=message)
            await self.journal.write_message(agent_message)
            self.logger.info(message)
            return None

    def create_function_tool(self, function_json: Dict[str, Any], name: str) -> BaseTool:
        """Create a function tool from JSON specification"""

        # In the case of external agents, if they report a name at all, they will
        # report something different that does not identify them as external.
        # Also, most internal agents do not have a name identifier on their functional
        # JSON, which is required.  Use the agent name we are using for look-up for that
        # regardless of intent.
        function_json["name"] = name

        return LangChainOpenAIFunctionTool.from_function_json(function_json, self.tool_caller)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""

import json
from typing import Any
from typing import Dict
from typing import Tuple

from http import HTTPStatus

from neuro_san.internals.interfaces.dictionary_validator import DictionaryValidator
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.service.http.handlers.base_request_handler import BaseRequestHandler
from neuro_san.service.utils.mcp_server_context import McpServerContext
from neuro_san.service.mcp.interfaces.client_session_policy import ClientSessionPolicy
from neuro_san.service.mcp.interfaces.client_session_policy import MCP_SESSION_ID, MCP_PROTOCOL_VERSION
from neuro_san.service.mcp.util.mcp_errors_util import McpErrorsUtil
from neuro_san.service.mcp.validation.tool_request_validator import ToolRequestValidator
from neuro_san.service.mcp.mcp_errors import McpError
from neuro_san.service.mcp.processors.mcp_tools_processor import McpToolsProcessor
from neuro_san.service.mcp.processors.mcp_resources_processor import McpResourcesProcessor
from neuro_san.service.mcp.processors.mcp_prompts_processor import McpPromptsProcessor
from neuro_san.service.mcp.processors.mcp_initialize_processor import McpInitializeProcessor
from neuro_san.service.mcp.processors.mcp_ping_processor import McpPingProcessor


class McpRootHandler(BaseRequestHandler):
    """
    Class implementing top-level MCP request handler.
    Note that since /mcp is our single MCP endpoint,
    all MCP requests are handled by this class.
    """
    # pylint: disable=attribute-defined-outside-init
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def initialize(self, **kwargs):
        """
        This method is called by Tornado framework to allow
        injecting service-specific data into local handler context.
        """
        # Initialize members of the base class BaseRequestHandler:
        super().initialize(**kwargs)

        # type: McpServerContext
        self.mcp_context: McpServerContext = self.server_context.get_mcp_server_context()
        # A dictionary of string (describing scope) to
        #     AgentNetworkStorage instance which keeps all the AgentNetwork instances
        #     of a particular grouping.
        self.network_storage_dict: Dict[str, AgentNetworkStorage] = self.server_context.get_network_storage_dict()

        # For tool requests, we need to validate tool call arguments:
        self.tool_request_validator: ToolRequestValidator = ToolRequestValidator(self.openapi_service_spec)

        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Methods", "GET, POST, DELETE, OPTIONS")
        headers: str = "Content-Type, Transfer-Encoding"
        metadata_headers: str = ", ".join(self.forwarded_request_metadata)
        if len(metadata_headers) > 0:
            headers += f", {metadata_headers}"
        # Set all allowed headers:
        self.set_header("Access-Control-Allow-Headers", headers)

    async def post(self):
        """
        Implementation of the top-level POST request handler for MCP call.
        """
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        metadata: Dict[str, Any] = self.get_metadata()
        request_id = "unknown"

        try:
            # Parse JSON body
            data = json.loads(self.request.body)

            # Validate incoming request content:
            request_validator: DictionaryValidator = self.mcp_context.get_request_validator()
            validation_errors = request_validator.validate(data)
            if validation_errors:
                extra_error: str = "; ".join(validation_errors)
                error_msg: Dict[str, Any] =\
                    McpErrorsUtil.get_protocol_error(request_id, McpError.InvalidRequest, extra_error)
                self.set_status(HTTPStatus.BAD_REQUEST)
                self.write(error_msg)
                self.logger.error(self.get_metadata(), f"Error: Invalid MCP request: {extra_error}")
                self.do_finish()
                return
        except json.JSONDecodeError as exc:
            error_msg: Dict[str, Any] =\
                McpErrorsUtil.get_protocol_error(
                    request_id,
                    McpError.ParseError,
                    "Invalid JSON format")
            self.set_status(HTTPStatus.BAD_REQUEST)
            self.write(error_msg)
            self.logger.error(self.get_metadata(), "error: Invalid JSON format: %s", str(exc))
            self.do_finish()
            return

        # We have valid MCP request:
        request_id = data.get("id", "absent")
        method: str = data.get("method")
        session_id: str = self.request.headers.get(MCP_SESSION_ID, None)
        protocol_version: str = self.request.headers.get(MCP_PROTOCOL_VERSION, None)
        if protocol_version:
            protocol_version = protocol_version.strip()

        # First check if we have handshake/initialize session requests -
        # these do not require valid protocol version or valid session:
        # MCP connection/session lifecycle is defined here:
        # https://modelcontextprotocol.io/specification/2025-06-18/basic/lifecycle
        new_session_id, request_done = await self.handle_handshake(method, data, session_id, request_id, metadata)
        if request_done:
            if new_session_id:
                self.set_header(MCP_SESSION_ID, new_session_id)
            self.do_finish()
            return

        # For all other methods, we need to have valid protocol version and valid session id,
        # it is possible that session id is not provided at all, i.e. None
        if protocol_version != self.mcp_context.get_protocol_version():
            extra_error: str = f"unsupported protocol version {protocol_version}"
            error_msg: Dict[str, Any] =\
                McpErrorsUtil.get_protocol_error(request_id, McpError.InvalidProtocolVersion, extra_error)
            self.set_status(HTTPStatus.BAD_REQUEST)
            self.write(error_msg)
            self.logger.error(self.get_metadata(), f"error: {extra_error}")
            self.do_finish()
            return
        session_policy: ClientSessionPolicy = self.mcp_context.get_session_policy()
        session_active = session_policy.is_session_active(session_id)
        if not session_active:
            extra_error: str = "invalid or inactive session id"
            error_msg: Dict[str, Any] =\
                McpErrorsUtil.get_protocol_error(request_id, McpError.InvalidSession, extra_error)
            self.set_status(HTTPStatus.UNAUTHORIZED)
            self.write(error_msg)
            self.logger.error(self.get_metadata(), f"error: {extra_error}")
            self.do_finish()
            return

        try:
            if method == "tools/list":
                tools_processor: McpToolsProcessor =\
                    McpToolsProcessor(
                        self.logger,
                        self.network_storage_dict,
                        self.agent_policy,
                        self.tool_request_validator)
                result_dict: Dict[str, Any] = await tools_processor.list_tools(request_id, metadata)
                self.set_status(HTTPStatus.OK)
                self.write(result_dict)
            elif method == "tools/call":
                tools_processor: McpToolsProcessor =\
                    McpToolsProcessor(
                        self.logger,
                        self.network_storage_dict,
                        self.agent_policy,
                        self.tool_request_validator)
                call_params: Dict[str, Any] = data.get("params", {})
                tool_name: str = call_params.get("name")
                call_args: Dict[str, Any] = call_params.get("arguments", {})
                # Validate tool arguments:
                validation_errors = self.tool_request_validator.validate(call_args)
                if validation_errors:
                    extra_error: str = "; ".join(validation_errors)
                    error_msg: Dict[str, Any] = \
                        McpErrorsUtil.get_protocol_error(request_id, McpError.InvalidRequest, extra_error)
                    self.set_status(HTTPStatus.BAD_REQUEST)
                    self.write(error_msg)
                    self.logger.error(self.get_metadata(), f"Error: Invalid tool call request: {extra_error}")
                    return

                prompt: Dict[str, Any] = call_args.get("user_message", {})
                chat_context: Dict[str, Any] = call_args.get("chat_context", None)
                chat_filter: Dict[str, Any] = call_args.get("chat_filter", None)
                sly_data: Dict[str, Any] = call_args.get("sly_data", None)
                result_dict: Dict[str, Any] =\
                    await tools_processor.call_tool(
                        request_id, metadata,
                        tool_name,
                        prompt,
                        chat_context,
                        chat_filter,
                        sly_data)
                self.set_status(HTTPStatus.OK)
                self.write(result_dict)
            elif method == "resources/list":
                resources_processor: McpResourcesProcessor = McpResourcesProcessor(self.logger)
                result_dict: Dict[str, Any] = await resources_processor.list_resources(request_id, metadata)
                self.set_status(HTTPStatus.OK)
                self.write(result_dict)
            elif method == "prompts/list":
                prompts_processor: McpPromptsProcessor = McpPromptsProcessor(self.logger)
                result_dict: Dict[str, Any] = await prompts_processor.list_prompts(request_id, metadata)
                self.set_status(HTTPStatus.OK)
                self.write(result_dict)
            else:
                # Method is not found/not supported
                extra_error: str = f"method {method} not found"
                error_msg: Dict[str, Any] =\
                    McpErrorsUtil.get_protocol_error(request_id, McpError.NoMethod, extra_error)
                self.set_status(HTTPStatus.BAD_REQUEST)
                self.write(error_msg)
                self.logger.error(self.get_metadata(), f"error: Method {method} not found")
        except Exception as exc:  # pylint: disable=broad-exception-caught
            error_msg: Dict[str, Any] =\
                McpErrorsUtil.get_protocol_error(
                    request_id,
                    McpError.ServerError,
                    f"exception during {method} handling")
            self.set_status(HTTPStatus.INTERNAL_SERVER_ERROR)
            self.write(error_msg)
            self.logger.error(self.get_metadata(), "error: Server error %s: %s", method, str(exc))
        finally:
            # We are done with response stream:
            self.do_finish()

    async def handle_handshake(
            self,
            method: str,
            request_data: Dict[str, Any],
            session_id: str,
            request_id,
            metadata: Dict[str, Any]) -> Tuple[str, bool]:
        """
        Handle handshake/initialize session request if request method is applicable.
        :param method: MCP method name
        :param request_data: MCP request data dictionary
        :param session_id: MCP client session id taken from request headers
        :request_id: MCP request id
        :param metadata: http-level request metadata;
        :return: tuple of 2 values:
                 first - new session id if it was generated here;
                         None otherwise
                second - True if request was handled here;
                         False otherwise
        """
        try:
            if method == "initialize":
                # Handle handshake/initialize session request
                handshake_processor: McpInitializeProcessor = McpInitializeProcessor(self.mcp_context, self.logger)
                result_dict, session_id =\
                    await handshake_processor.initialize_handshake(request_id, metadata, request_data["params"])
                if session_id is not None:
                    self.set_header(MCP_SESSION_ID, session_id)
                self.set_status(HTTPStatus.OK)
                self.write(result_dict)
                return session_id, True
            if method == "notifications/initialized":
                # Handle client acknowledgement of initialization response,
                # this activates the session on the server side for further operations.
                handshake_processor: McpInitializeProcessor = McpInitializeProcessor(self.mcp_context, self.logger)
                result: bool = await handshake_processor.activate_session(session_id, metadata)
                response_code: int = HTTPStatus.ACCEPTED if result else HTTPStatus.NOT_FOUND
                if session_id is not None:
                    self.set_header(MCP_SESSION_ID, session_id)
                self.set_status(response_code)
                # We do not have any response body for this request
                return None, True
            if method == "ping":
                # Handle client-side ping,
                # we don't care about sessions here.
                ping_processor: McpPingProcessor = McpPingProcessor(self.logger)
                _ = await ping_processor.ping(session_id, metadata)
                response_code: int = HTTPStatus.OK
                self.set_status(response_code)
                # We do not have any response body for this request
                return None, True
        except Exception as exc:  # pylint: disable=broad-exception-caught
            error_msg: Dict[str, Any] = \
                McpErrorsUtil.get_protocol_error(
                    request_id,
                    McpError.ServerError,
                    f"exception during {method} handling")
            self.set_status(HTTPStatus.INTERNAL_SERVER_ERROR)
            self.write(error_msg)
            self.logger.error(self.get_metadata(), "error: Server error %s: %s", method, str(exc))
            return None, True
        return None, False

    async def delete(self):
        """
        Implementation of top-level DELETE request handler for MCP call.
        """

        metadata: Dict[str, Any] = self.get_metadata()
        request_id = "unknown"

        # We only expect MCP client session id taken from request headers:
        session_id: str = self.request.headers.get(MCP_SESSION_ID, None)

        request_status: int = HTTPStatus.NO_CONTENT
        if session_id is not None:
            session_policy: ClientSessionPolicy = self.mcp_context.get_session_policy()
            deleted: bool = session_policy.delete_session(session_id)
            if deleted:
                self.logger.info(metadata, "Session %s deleted by client", session_id)
            else:
                extra_error: str = "Session id not found"
                error_msg: Dict[str, Any] =\
                    McpErrorsUtil.get_protocol_error(request_id, McpError.InvalidSession, extra_error)
                self.set_status(HTTPStatus.NOT_FOUND)
                self.write(error_msg)
                self.logger.error(metadata, f"Error: {extra_error}")
        else:
            # No session id is provided in this request:
            # report bad request
            request_status = HTTPStatus.UNAUTHORIZED
        self.set_status(request_status)
        self.do_finish()

    async def get(self):
        """
        Implementation of top-level GET request handler for MCP call.
        """
        # Consider GET request for MCP endpoint to be a service health check
        self.set_status(HTTPStatus.OK)
        self.do_finish()




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

from logging import getLogger
from logging import Logger

from pathlib import Path

from watchdog.observers.polling import PollingObserver

from neuro_san.service.watcher.registries.registry_change_handler import RegistryChangeHandler
from neuro_san.service.watcher.registries.registry_observer import RegistryObserver


class PollingRegistryObserver(RegistryObserver):
    """
    Observer class for manifest file and its directory.
    """

    def __init__(self, manifest_path: Union[str, List[str]], poll_seconds: int):

        self.manifest_path: Union[str, List[str]] = manifest_path
        self.registry_observers: Dict[str, PollingObserver] = {}
        self.logger: Logger = getLogger(self.__class__.__name__)
        self.poll_seconds: int = poll_seconds
        self.event_handler: RegistryChangeHandler = RegistryChangeHandler()

        split_manifest_paths: List[str] = manifest_path
        if isinstance(manifest_path, str):
            split_manifest_paths = manifest_path.split(" ")

        for split_component in split_manifest_paths:
            one_manifest_path: str = Path(split_component)
            registry_path: str = str(one_manifest_path.parent)
            if registry_path not in self.registry_observers:
                # One observer per unique registry path
                self.registry_observers[registry_path] = PollingObserver(timeout=self.poll_seconds)

    def start(self):
        """
        Start running observer
        """
        for registry_path, observer in self.registry_observers.items():
            observer.schedule(self.event_handler, path=registry_path, recursive=True)
            observer.start()
            self.logger.info("Registry polling watchdog started on: %s for manifest %s with polling every %d sec",
                             registry_path, self.manifest_path, self.poll_seconds)

    def reset_event_counters(self) -> Tuple[int, int, int]:
        """
        Reset event counters and return current counters.
        """
        return self.event_handler.reset_event_counters()




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Generator

from neuro_san.interfaces.agent_session_constants import AgentSessionConstants


class AsyncAgentSession(AgentSessionConstants):
    """
    Asynchronous interface for the initiation and continuity of a single Agent session.
    """

    async def function(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the FunctionRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the FunctionResponse
                    protobufs structure. Has the following keys:
                "function" - the dictionary description of the function
        """
        raise NotImplementedError

    async def connectivity(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the ConnectivityRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the ConnectivityResponse
                    protobufs structure. Has the following keys:
                "connectivity_info" - the list of connectivity descriptions for
                                    each node in the agent network the service
                                    wants the client ot know about.
        """
        raise NotImplementedError

    async def streaming_chat(self, request_dict: Dict[str, Any]) -> Generator[Dict[str, Any], None, None]:
        """
        :param request_dict: A dictionary version of the ChatRequest
                    protobufs structure. Has the following keys:
            "user_message" - A ChatMessage dict representing the user input to the chat stream
            "chat_context" - A ChatContext dict representing the state of the previous conversation
                            (if any)

        :return: An iterator of dictionary versions of the ChatResponse
                    protobufs structure. Has the following keys:
            "response"      - An optional ChatMessage dictionary.  See chat.proto for details.

            Note that responses to the chat input might be numerous and will come as they
            are produced until the system decides there are no more messages to be sent.
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Union

from langchain_core.tools import BaseTool


class ContextTypeToolboxFactory:
    """
    Interface for factory classes that create tools or toolkits from a toolbox.

    Most methods accept a configuration dictionary, where each key is a tool name, and each value is
    a dictionary containing the corresponding tool's setup information. The configuration dictionary
    supports the following keys for each tool.

    Langchain's Tool:
        - "class":  The class of the tool or toolkit.
                    This key is required. A ValueError will be raised if not provided.

        - "args":   A dictionary of constructor or class method arguments used to instantiate the tool
                    or toolkit.

    Coded Tool:
        - "class":  Module and class in the format of tool_module.ClassName.

        - "description":  When and how to use the tool.

        - "parameters":  Information on arguments of the tool.
    """

    def load(self):
        """
        Goes through the process of loading any user extensions and/or configuration
        files
        """
        raise NotImplementedError

    def create_tool_from_toolbox(
            self,
            tool_name: str,
            user_args: Dict[str, Any]
    ) -> Union[BaseTool, Dict[str, Any], List[BaseTool]]:
        """
        Create a tool instance from the fully-specified tool config.
        :param tool_name: The name of the tool to instantiate.
        :param user_args: Arguments provided by the user, which override the args in config file.
        :return: A tool instance native to the context type.
                Can raise a ValueError if the config's class or tool_name value is
                unknown to this method.
        """
        raise NotImplementedError

    def get_shared_coded_tool_class(self, tool_name: str) -> str:
        """
        Get class of the shared coded tool from toolbox

        :param tool_name: The name of the tool
        :return: The class of the coded tool from toolbox
        """
        raise NotImplementedError

    def get_tool_info(self, tool_name: str) -> Dict[str, Any]:
        """
        :param tool_name: The name of the tool.
        :return: The toolbox dictionary entry for the tool name
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from langchain_core.language_models.base import BaseLanguageModel

from neuro_san.internals.run_context.langchain.llms.llm_policy import LlmPolicy


class LangChainLlmResources:
    """
    Class for representing a LangChain model
    together with run-time policy necessary for model usage by the service.
    """

    def __init__(self, model: BaseLanguageModel, llm_policy: LlmPolicy = None):
        """
        Constructor.
        :param model: Language model used.
        :param llm_policy: optional LlmPolicy object to manage connections to LLM host.
        """
        self.model: BaseLanguageModel = model
        self.llm_policy: LlmPolicy = llm_policy

    def get_model(self) -> BaseLanguageModel:
        """
        :return: the BaseLanguageModel
        """
        return self.model

    def get_llm_policy(self) -> LlmPolicy:
        """
        :return: the LlmPolicy used by the model
        """
        return self.llm_policy

    async def delete_resources(self):
        """
        Release the run-time resources used by the model
        """
        if self.llm_policy is not None:
            await self.llm_policy.delete_resources()




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Union

from copy import copy

from neuro_san.internals.messages.traced_message import TracedMessage


class AgentFrameworkMessage(TracedMessage):
    """
    TracedMessage implementation of a message from the agent framework
    """

    structure: Optional[Dict[str, Any]] = None
    sly_data: Optional[Dict[str, Any]] = None
    chat_context: Optional[Dict[str, Any]] = None

    type: Literal["agent-framework"] = "agent-framework"

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def __init__(self, content: Union[str, List[Union[str, Dict]]] = "",
                 chat_context: Dict[str, Any] = None,
                 sly_data: Dict[str, Any] = None,
                 structure: Dict[str, Any] = None,
                 trace_source: AgentFrameworkMessage = None,
                 **kwargs: Any) -> None:
        """
        Pass in content as positional arg.

        :param content: The string contents of the message.
        :param chat_context: A dictionary that fully desbribes the state of play
                    of the chat conversation such that when it is passed on to a
                    different server, the conversation can continue uninterrupted.
        :param sly_data: A dictionary of private data, separate from the chat stream.
        :param structure: A dictionary previously extracted from the content
                        that had been optionally detected by the system as JSON text.
                        The idea is to have the server do the hard parsing so the
                        multitude of clients do not have to rediscover how to best do it.
        :param trace_source: A message of the same type to prepare for tracing display
        :param kwargs: Additional fields to pass to the superclass
        """
        super().__init__(content=content, trace_source=trace_source, **kwargs)
        self.chat_context: Dict[str, Any] = chat_context
        self.sly_data: Dict[str, Any] = sly_data
        self.structure: Dict[str, Any] = structure

    @property
    def lc_kwargs(self) -> Dict[str, Any]:
        """
        :return: the keyword arguments for serialization.
        """
        return {
            "content": self.content,
            "structure": self.structure,
            "sly_data": self.sly_data,
            "chat_context": self.chat_context,
        }

    def translate_for_trace(self, key: str, value: Any) -> Tuple[str, Any]:
        """
        :param key: The key to consider/translate.
        :param value: The value to consider/translate
        :return: A tuple with the new key and new value to be shown in the trace.
                New keys that are None are not included in the additional_kwargs.
                The default implementation simply ensures that there is something in the
                value to trace display to maximize information.
        """
        new_key, new_value = super().translate_for_trace(key, value)
        if not new_key:
            return None, None

        # Specifically redact any sly_data values, but keep the keys.
        # The intent here is to not transmit any sensitive information
        # that might make it to some other host.
        if new_key == "sly_data":
            # Shallow copy the original sly_data dictionary
            new_value = copy(value)
            for sly_data_key in new_value:
                # Keep the keys but redact the values
                new_value[sly_data_key] = "<redacted>"

        return new_key, new_value




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Tuple
from typing import Type
from typing import Union

import copy
import traceback

from logging import getLogger
from logging import Logger
from inspect import iscoroutinefunction

from langchain_core.messages.base import BaseMessage

from leaf_common.config.resolver_util import ResolverUtil

from neuro_san.interfaces.reservationist import Reservationist
from neuro_san.internals.chat.async_collating_queue import AsyncCollatingQueue
from neuro_san.internals.chat.chat_history_message_processor import ChatHistoryMessageProcessor
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.graph.registry.agent_tool_registry import AgentToolRegistry
from neuro_san.internals.graph.activations.sly_data_redactor import SlyDataRedactor
from neuro_san.internals.interfaces.context_type_tracing_context_factory import ContextTypeTracingContextFactory
from neuro_san.internals.interfaces.front_man import FrontMan
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.interfaces.run_target import RunTarget
from neuro_san.internals.journals.intercepting_journal import InterceptingJournal
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.agent_framework_message import AgentFrameworkMessage
from neuro_san.internals.messages.base_message_dictionary_converter import BaseMessageDictionaryConverter
from neuro_san.internals.run_context.factory.run_context_factory import RunContextFactory
from neuro_san.internals.run_context.factory.master_tracing_context_factory import MasterTracingContextFactory
from neuro_san.internals.run_context.interfaces.run_context import RunContext
from neuro_san.message_processing.message_processor import MessageProcessor
from neuro_san.message_processing.answer_message_processor import AnswerMessageProcessor
from neuro_san.message_processing.structure_message_processor import StructureMessageProcessor

# Lazily import specific errors from llm providers
PATIENCE_ERRORS: Tuple[Type[Any], ...] = ResolverUtil.create_type_tuple([
                                            "openai.BadRequestError",
                                         ])


# pylint: disable=too-many-instance-attributes
class DataDrivenChatSession(RunTarget):
    """
    ChatSession implementation that consolidates policy
    in using data-driven agent tool graphs.
    """

    def __init__(self, agent_network: AgentNetwork):
        """
        Constructor

        :param agent_network: The AgentNetwork to use.
        """
        # We make a copy of the AgentNetwork at this level to allow for interactions
        # within this scope to modify the network topology.
        agent_network_copy: AgentNetwork = copy.deepcopy(agent_network)
        self.registry: AgentToolRegistry = AgentToolRegistry(agent_network_copy)

        self.front_man: FrontMan = None
        self.sly_data: Dict[str, Any] = {}

        # This is the context policy container that pertains to the invocation
        # of run_it()
        self.invocation_context: InvocationContext = None
        self.interceptor: InterceptingJournal = None
        self.original_input_message: AgentFrameworkMessage = None
        self.last_message_sent: bool = False

    async def set_up(self, invocation_context: InvocationContext,
                     sly_data: Dict[str, Any] = None,
                     chat_context: Dict[str, Any] = None):
        """
        Resets or sets the instance up for the first time.
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be None.
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
        """

        # Reset any sly data
        # This ends up being the one reference to the sly_data that gets passed around
        # to the graph components. Updating this updates everyone else.
        self.sly_data = {}

        # Reset what we might have created before.
        await self.delete_resources()

        # Update sly data (if any) with what was passed in.
        # Note that since this instance is the owner of the sly_data,
        # any update here should get transmitted to all the other graph components
        # because it is expected they share the reference and only interact with it
        # in a read-only fashion.
        if sly_data is not None:
            self.sly_data.update(sly_data)

        # Update sly data with metadata from the request headers so this information
        # can be made available in CodedTools, privately, without any leakage to LLMs.
        # Only add this if it doesn't already exist w/rt what the user gave us.
        if self.sly_data.get("request_metadata") is None:
            # Make a deep copy so the server's idea of the original metadata doesn't get modified
            # should any CodedTool somehow get the idea to modify it.
            self.sly_data["request_metadata"] = copy.deepcopy(invocation_context.get_metadata())

        run_context: RunContext = RunContextFactory.create_run_context(None, None,
                                                                       invocation_context=invocation_context,
                                                                       chat_context=chat_context)

        self.front_man = self.registry.create_front_man(self.sly_data, run_context)

        await self.front_man.create_any_resources()

    async def chat(self, user_input: str,
                   invocation_context: InvocationContext,
                   sly_data: Dict[str, Any] = None) -> Iterator[Dict[str, Any]]:
        """
        Performs a portion of the streaming_chat() responsibilities

        :param user_input: A string with the user's input
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be None.
        :return: An Iterator over dictionary representation of chat messages.
                The keys/values/structure of these chat message dictionaries will reflect
                instances of ChatMessage from chat.proto.

                Note that Iterators themselves are *not* simply lists. They are a Python
                construct intended for use in a for-loop that is allowed to come up with
                its content dynamically.  For our purposes, when an initiator of chat()
                gets a handle to this Iterator, they can begin looping/waiting on its contents
                without the content itself having been created yet.  This is a building
                block of streaming results even though direct callers may not actually
                be streaming.
        """
        if self.front_man is None:
            await self.set_up(invocation_context, sly_data)
        else:
            self.front_man.update_invocation_context(invocation_context)

        try:
            # DEF - drill further down for iterator from here to enable getting
            #       messages from downstream agents.
            raw_messages: List[BaseMessage] = await self.front_man.submit_message(user_input)

        except PATIENCE_ERRORS:
            # This can happen if the user is trying to send a new message
            # while it is still working on a previous message that has not
            # yet returned.
            raw_messages: List[BaseMessage] = [
                AgentFrameworkMessage(content="Patience, please. I'm working on it.")
            ]

            logger: Logger = getLogger(self.__class__.__name__)
            logger.error(traceback.format_exc())

        converter = BaseMessageDictionaryConverter(origin=self.front_man.get_origin())
        chat_messages: List[Dict[str, Any]] = []
        for raw_message in raw_messages:
            chat_message: Dict[str, Any] = converter.to_dict(raw_message)
            chat_messages.append(chat_message)

        return iter(chat_messages)

    # pylint: disable=too-many-locals
    async def streaming_chat(self, user_input: str,
                             invocation_context: InvocationContext,
                             sly_data: Dict[str, Any] = None,
                             chat_context: Dict[str, Any] = None):
        """
        Main streaming entry-point method for accepting new user input
        invoked by DirectAgentSession and its friends.

        :param user_input: A string with the user's input
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be None.
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
        :return: Nothing.  Response values are put on a queue whose consumtion is
                managed by the Iterator aspect of AsyncCollatingQueue on the InvocationContext.
        """
        # Set up some member variable state so that run_it() can use it
        self.invocation_context = invocation_context
        journal: Journal = self.invocation_context.get_journal()
        self.interceptor = InterceptingJournal(journal, origin=None)

        # Set up an input message that will show up in an Observability/tracing app
        # We keep this guy around for the duration of the chat session so that this class
        # has a pristine copy of what the inputs for use within run_it().
        self.original_input_message = AgentFrameworkMessage(content=user_input,
                                                            chat_context=chat_context,
                                                            sly_data=sly_data)
        # Make a copy of the input message for use in the tracing context
        # We can't use the same object as original_input_message because
        # the tracing infrastructure ends up transforming the message for display.
        input_message_for_show = AgentFrameworkMessage(trace_source=self.original_input_message)

        # Set up configuration for creating the tracing context.
        # These are the bare minimum required to get output and metadata correct
        # in the tracing reporting.
        config: Dict[str, Any] = {
            "interceptor": self.interceptor,
            "invocation_context": self.invocation_context
        }

        # Get a toolkit-specific implementation for creating the tracing context
        tracing_factory: ContextTypeTracingContextFactory = \
            MasterTracingContextFactory.create_tracing_context_factory()
        # For the factory args, we are our own run_target.
        tracing_context: RunTarget = tracing_factory.create_tracing_context(config, run_target=self)

        # Run the run_target that was given back by the factory.
        await tracing_context.run_it(input_message_for_show)

    async def run_it(self, inputs: AgentFrameworkMessage) -> AgentFrameworkMessage:
        """
        This method is effectively a callback which is invoked within
        the tracing context infrastructure for the toolkit/context_type.

        :param inputs: An AgentFrameworkMessage populated with the user's input.
        :return: The user input. (Outputs are handled by the tracing context infrastructure.)
        """
        # Get all our real input values from the original_input_message.
        # If we got it from the inputs arg, we'd get the for-show message
        # which has fields rearranged and even redacted.
        user_input: str = self.original_input_message.content
        sly_data: Dict[str, Any] = self.original_input_message.sly_data
        chat_context: Dict[str, Any] = self.original_input_message.chat_context

        if self.front_man is None:
            try:
                await self.set_up(self.invocation_context, sly_data, chat_context)
            except ValueError as exc:
                # This can happen if we have problems with LLM clients API keys:
                # Construct a message to send back to the client with the error information.
                message = AgentFrameworkMessage(content=str(exc))
                await self.finalize_run(message)
                return message

        # If we are invoked as an event, tell the caller that it's OK to disconnect early.
        # By definition, they are not expecting a custom response.
        if self.invocation_context.get_effective_invocation() == "event":
            empty: Dict[str, Any] = {}
            event_acknowledge = AgentFrameworkMessage(content="Event acknowledged", chat_context=empty)
            await self.finalize_run(event_acknowledge)

        # Actually run the chat and save information about it
        chat_messages: Iterator[Dict[str, Any]] = await self.chat(user_input, self.invocation_context, sly_data)
        message_list: List[Dict[str, Any]] = list(chat_messages)

        # Determine the chat_context to enable continuing the conversation
        return_chat_context: Dict[str, Any] = self.prepare_chat_context(message_list)

        # Get the front man spec. We will need it later for a few things.
        front_man_spec: Dict[str, Any] = self.front_man.get_agent_tool_spec()

        # Get the formats we should parse from the final answer from the config for the network.
        # As of 6/24/25, this is an unadvertised experimental feature.
        structure_formats: Union[str, List[str]] = front_man_spec.get("structure_formats")

        # Find "the answer" and have that be the content of the last message we send
        answer_processor = AnswerMessageProcessor(structure_formats=structure_formats)
        answer_processor.process_messages(message_list)
        answer: str = answer_processor.get_answer()
        if answer is None:
            # Can't have content as None or problems arise.
            answer = ""
        structure: Dict[str, Any] = answer_processor.get_structure()

        # Send back sly_data as the front-man permits
        redactor = SlyDataRedactor(front_man_spec,
                                   config_keys=["allow.to_upstream.sly_data"],
                                   allow_empty_dict=False)
        return_sly_data: Dict[str, Any] = redactor.filter_config(self.sly_data)

        # Stream over chat state as the last message
        # Use the interceptor to write the message.
        # This guy wraps the journal from the invocation context and listens to the
        # messages coming across, which allows us to report to the tracing infrastructure
        # at the end of the tracing context.
        message = AgentFrameworkMessage(content=answer, chat_context=return_chat_context,
                                        sly_data=return_sly_data, structure=structure)
        await self.finalize_run(message)

        # Bogus output, but need something for interface
        outputs: AgentFrameworkMessage = inputs
        return outputs

    async def finalize_run(self, message: BaseMessage):
        """
        This is a method that publishes the resulting message of the run
        and performs necessary finalization steps for run resources.

        :param message: The final message delivered from the run.
        :return: Nothing.
        """
        if self.last_message_sent:
            return
        self.last_message_sent = True

        # Use the interceptor to write the final message.
        # This guy wraps the journal from the invocation context and listens to the
        # messages coming across, which allows us to report to the tracing infrastructure
        # at the end of the tracing context.
        await self.interceptor.write_message(message, origin=None)

        # Put an end-marker on the queue to tell the consumer we truly are done
        # and it doesn't need to wait for any more messages.
        # The consumer await-s for queue.get()
        queue: AsyncCollatingQueue = self.invocation_context.get_queue()

        # The synchronous=True is necessary when an async HTTP request is at the get()-ing end of the queue,
        # as the journal messages come from inside a separate event loop from that request. The lock
        # taken here ends up being harmless in the synchronous request case (like for gRPC) because
        # we would only be blocking our own event loop.
        await queue.put_final_item(synchronous=True)

        # Now that we are done, tell the Reservationist that we used for this request
        # that there will be no more Reservations to corral.
        reservationist: Reservationist = self.invocation_context.get_reservationist()
        if reservationist is not None:
            await reservationist.close()

        # Close any objects on sly data that can be closed.
        await self.close_sly_data()

    async def delete_resources(self):
        """
        Frees up any service-side resources.
        """
        if self.front_man is not None:
            await self.front_man.delete_any_resources()
            self.front_man = None

    def prepare_chat_context(self, chat_message_history: List[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Prepare the chat context.

        :param chat_message_history: A list of ChatMessage dictionaries that
                comprise the front man's full chat history.
        :return: A ChatContext dictionary comprising the full state of play of
                the conversation such that it could be taken up on a different
                server instance
        """
        # OK if this is None
        max_message_history: int = self.front_man.get_agent_tool_spec().get("max_message_history")
        processor: MessageProcessor = ChatHistoryMessageProcessor(max_message_history)
        processor.process_messages(chat_message_history)

        chat_history: Dict[str, Any] = {
            "origin": self.front_man.get_origin(),
            "messages": processor.get_message_history()
        }

        # For now, we only send the front man's chat history, as that is the
        # state we had been preserving since the early days.  It is conceivable
        # this could expand to other agents in the hierarchy at some point.
        chat_context: Dict[str, Any] = {
            "chat_histories": [chat_history]
        }

        return chat_context

    def create_outgoing_message_processor(self) -> MessageProcessor:
        """
        :return: A MessageProcessor that filters messages outgoing to the client.
                How this works is based on settings on the front man.
                Can be None.
        """
        message_processor: MessageProcessor = None

        front_man_name: str = self.registry.find_front_man()
        front_man_spec: Dict[str, Any] = self.registry.get_agent_tool_spec(front_man_name)

        # Get the formats we should parse from the final answer from the config for the network.
        # As of 6/24/25, this is an unadvertised experimental feature.
        structure_formats: Union[str, List[str]] = front_man_spec.get("structure_formats")
        if structure_formats is None:
            return message_processor

        # Eventually this might be a CompositeMessageProcessor
        message_processor = StructureMessageProcessor(structure_formats)
        return message_processor

    async def close_sly_data(self):
        """
        Close any sly data value that can be closed.
        """
        if self.sly_data is None or len(self.sly_data) == 0:
            # Nothing to close
            return

        # Try closing anything that can be close()-d in the sly_data.
        for key, value in self.sly_data.items():
            # Note: It's possible we might want to try closing keys as well
            description: str = f"sly_data[{key}]"
            await self.close_one(value, description)

    async def close_one(self, value: Any, description: str):
        """
        Close a single object if it can be closed.
        :param value: The object to close
        :param description: Where the object to close came from
        """
        if not hasattr(value, "close"):
            # Nothing to close
            return

        close_method = getattr(value, "close")
        if not callable(close_method):
            # No method to call
            return

        # Call the close method
        try:
            if iscoroutinefunction(close_method):
                await close_method()
            else:
                close_method()

        except TypeError:
            # Not enough arguments to call close().
            logger: Logger = getLogger(self.__class__.__name__)
            logger.error("Unable to close() %s", description)
            logger.error(traceback.format_exc())




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.run_context.interfaces.agent_spec_provider import AgentSpecProvider
from neuro_san.internals.run_context.interfaces.agent_network_inspector import AgentNetworkInspector
from neuro_san.internals.run_context.interfaces.run import Run


class ToolCaller(AgentSpecProvider):
    """
    Interface for Tools that call Agents/LLMs as functions.
    This is called by langchain Tools and implemented by CallingTool.
    """

    async def make_tool_function_calls(self, component_run: Run) -> Run:
        """
        Calls all of the callable_components' functions

        :param component_run: The Run which the component is operating under
        :return: A potentially updated Run for the component
        """
        raise NotImplementedError

    def get_inspector(self) -> AgentNetworkInspector:
        """
        :return: The AgentNetworkInspector that contains the specs of all the tools
        """
        raise NotImplementedError

    def get_agent_tool_spec(self) -> Dict[str, Any]:
        """
        :return: the dictionary describing the data-driven agent
        """
        raise NotImplementedError

    def get_name(self) -> str:
        """
        :return: the name of the data-driven agent as it comes from the spec
        """
        raise NotImplementedError

    def get_sly_data(self) -> Dict[str, Any]:
        """
        :return: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be an empty dictionary.
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from logging import getLogger
from logging import Logger

from leaf_common.config.config_filter import ConfigFilter


class ManifestKeyConfigFilter(ConfigFilter):
    """
    Implementation of the ConfigFilter interface that reads the contents
    of a single manifest file for agent networks/registries, converting
    keys to a standardized form.
    """

    def __init__(self, manifest_file: str):
        """
        Constructor

        :param manifest_file: The name of the manifest file we are processing for logging purposes
        """
        self.logger: Logger = getLogger(self.__class__.__name__)
        self.manifest_file: str = manifest_file

    def filter_config(self, basis_config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Filters the given basis config.

        :param basis_config: The config dictionary to act as the basis
                for filtering
        :return: A config dictionary, potentially modified as per the
                policy encapsulated by the implementation
        """

        filtered: Dict[str, Any] = {}

        for key, value in basis_config.items():

            # Key here is an agent name in a form that we choose.
            # Keys sometimes come with quotes.
            manifest_key: str = key.replace(r'"', "")
            manifest_key = manifest_key.strip()

            filtered[manifest_key] = value

        return filtered




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from contextlib import suppress

from langchain_core.language_models.base import BaseLanguageModel

from neuro_san.internals.run_context.langchain.llms.llm_policy import LlmPolicy


class OllamaLlmPolicy(LlmPolicy):
    """
    LlmPolicy implementation for Ollama.

    Ollama models do not allow for passing in an externally managed web client.
    """

    def create_llm(self, config: Dict[str, Any], model_name: str, client: Any) -> BaseLanguageModel:
        """
        Create a BaseLanguageModel instance from the fully-specified llm config
        for the llm class that the implementation supports.  Chat models are usually
        per-provider, where the specific model itself is an argument to its constructor.

        :param config: The fully specified llm config
        :param model_name: The name of the model
        :param client: The web client to use (if any)
        :return: A BaseLanguageModel (can be Chat or LLM)
        """
        # Use lazy loading to prevent installing the world
        # pylint: disable=invalid-name
        ChatOllama = self.resolver.resolve_class_in_module("ChatOllama",
                                                           module_name="langchain_ollama",
                                                           install_if_missing="langchain-ollama")
        # Higher temperature is more random
        llm = ChatOllama(
            model=model_name,
            mirostat=config.get("mirostat"),
            mirostat_eta=config.get("mirostat_eta"),
            mirostat_tau=config.get("mirostat_tau"),
            num_ctx=config.get("num_ctx"),
            num_gpu=config.get("num_gpu"),
            num_thread=config.get("num_thread"),
            num_predict=config.get("num_predict", config.get("max_tokens")),
            reasoning=config.get("reasoning"),
            repeat_last_n=config.get("repeat_last_n"),
            repeat_penalty=config.get("repeat_penalty"),
            temperature=config.get("temperature"),
            seed=config.get("seed"),
            stop=config.get("stop"),
            tfs_z=config.get("tfs_z"),
            top_k=config.get("top_k"),
            top_p=config.get("top_p"),
            keep_alive=config.get("keep_alive"),
            base_url=config.get("base_url"),

            # If omitted, this defaults to the global verbose value,
            # accessible via langchain_core.globals.get_verbose():
            # https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/globals.py#L53
            #
            # However, accessing the global verbose value during concurrent initialization
            # can trigger the following warning:
            #
            # UserWarning: Importing verbose from langchain root module is no longer supported.
            # Please use langchain.globals.set_verbose() / langchain.globals.get_verbose() instead.
            # old_verbose = langchain.verbose
            #
            # To prevent this, we explicitly set verbose=False here (which matches the default
            # global verbose value) so that the warning is never triggered.
            verbose=False,
        )
        return llm

    async def delete_resources(self):
        """
        Release the run-time resources used by the model
        """
        if self.llm is None:
            return

        # Do the necessary reach-ins to successfully shut down the web client

        # This is really an ollama.AsyncClient, but we don't really want to do the Resolver here.
        # Note we don't want to do this in the constructor, as OllamaChat lazily
        # creates these as needed via a private member that needs to be done in its own time
        # via Ollama infrastructure.  By the time we get here, it's already been created.
        ollama_async_client: Any = self.llm._async_client       # pylint:disable=protected-access

        if ollama_async_client is not None:
            with suppress(Exception):
                await ollama_async_client._client.aclose()      # pylint:disable=protected-access

        # Let's not do this again, shall we?
        self.llm = None




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Dict

from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider


class SingleAgentNetworkProvider(AgentNetworkProvider):
    """
    Class providing current AgentNetwork for a given agent in the service scope.
    """
    def __init__(self, agent_name: str, agents_table: Dict[str, AgentNetwork]):
        """
        Constructor.
        :param agent_name: name of an agent to provide AgentNetwork instances for;
        :param agents_table: service-wide table mapping agent names to their
            currently active AgentNetwork instances.
            This table is assumed to be dynamically modified outside a single agent scope.
        """
        self.agent_name = agent_name
        self.agents_table: Dict[str, AgentNetwork] = agents_table

    def get_agent_network(self) -> AgentNetwork:
        """
        :return: Current AgentNetwork instance for specific agent name.
                None if this does not exist for the instance's agent_name.
        """
        return self.agents_table.get(self.agent_name)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Dict

from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.graph.persistence.registry_manifest_restorer import RegistryManifestRestorer
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage


class DirectAgentStorageUtil:
    """
    Sets up AgentNetworkStorage for direct usage.
    """

    @staticmethod
    def create_network_storage(manifest_networks: Dict[str, Dict[str, AgentNetwork]] = None,
                               storage_type: str = "public") -> AgentNetworkStorage:
        """
        Creates an AgentNetworkStorage instance for a given type.

        :param manifest_networks: Optional structure that is handed back from a RegistryManifestRestorer.restore()
                        call.  This has major keys being different network storage options like
                        "public" and "protected". The values are agent name -> AgentNetwork mappings.
                        By default the value is None, indicating we need to get this information
                        by calling the RegistryManifestRestorer.
        :param storage_type: The type of storage ("public" or "protected")
                        Default value is "public".
        :return: An AgentNetworkStorage populated from the Registry Manifest
        """
        network_storage = AgentNetworkStorage()

        if manifest_networks is None:
            manifest_restorer = RegistryManifestRestorer()
            manifest_networks = manifest_restorer.restore()

        storage_networks: Dict[str, AgentNetwork] = manifest_networks.get(storage_type)
        if storage_networks is None:
            return None

        for agent_name, agent_network in storage_networks.items():
            network_storage.add_agent_network(agent_name, agent_network)

        return network_storage




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import AsyncGenerator
from typing import List

from asyncio import Future
from contextlib import suppress
from copy import copy
import logging

from leaf_common.asyncio.asyncio_executor import AsyncioExecutor
from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san.interfaces.async_agent_session import AsyncAgentSession
from neuro_san.internals.chat.connectivity_reporter import ConnectivityReporter
from neuro_san.internals.chat.data_driven_chat_session import DataDrivenChatSession
from neuro_san.internals.filters.message_filter import MessageFilter
from neuro_san.internals.filters.message_filter_factory import MessageFilterFactory
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.message_processing.message_processor import MessageProcessor
from neuro_san.session.session_invocation_context import SessionInvocationContext


class AsyncDirectAgentSession(AsyncAgentSession):
    """
    Direct guts for an AsyncAgentSession.
    """

    LOG_AGENT_NETWORK_USAGE: bool = True

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self,
                 agent_network: AgentNetwork,
                 invocation_context: SessionInvocationContext,
                 metadata: Dict[str, Any] = None,
                 security_cfg: Dict[str, Any] = None):
        """
        Constructor

        :param agent_network: The AgentNetwork to use for the session.
        :param invocation_context: The SessionInvocationContext to use to consult
                        for policy objects scoped at the invocation level.
        :param metadata: A dictionary of request metadata to be forwarded
                        to subsequent yet-to-be-made requests.
        :param security_cfg: A dictionary of parameters used to
                        secure the TLS and the authentication of the gRPC
                        connection.  Supplying this implies use of a secure
                        GRPC Channel.  If None, uses insecure channel.
        """
        # These aren't used yet
        self._metadata: Dict[str, Any] = metadata
        self._security_cfg: Dict[str, Any] = security_cfg

        self.invocation_context: SessionInvocationContext = invocation_context
        self.agent_network: AgentNetwork = agent_network
        self.request_id: str = None
        if metadata is not None:
            self.request_id = metadata.get("request_id")
        self.logger = logging.getLogger(self.__class__.__name__)

    def _log_agent_network_usage(self, operation: str):
        """
        Logs the usage of the agent network if enabled.
        :param operation: The operation being performed (e.g., "function", "connectivity", "streaming_chat").
        """
        if self.LOG_AGENT_NETWORK_USAGE:
            self.logger.info("Agent network '%s' (ID: %d) used for operation: %s",
                             self.agent_network.get_network_name(),
                             id(self.agent_network),
                             operation)

    async def function(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the FunctionRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the FunctionResponse
                    protobufs structure. Has the following keys:
                "function" - the dictionary description of the function
        """
        self._log_agent_network_usage("function")
        _ = request_dict
        response_dict: Dict[str, Any] = {
        }

        front_man: str = self.agent_network.find_front_man()
        if front_man is not None:
            spec: Dict[str, Any] = self.agent_network.get_agent_tool_spec(front_man)
            empty: Dict[str, Any] = {}
            function: Dict[str, Any] = spec.get("function", empty)
            response_dict = {
                "function": function,
            }

        return response_dict

    async def connectivity(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the ConnectivityRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the ConnectivityResponse
                    protobufs structure. Has the following keys:
                "connectivity_info" - the list of connectivity descriptions for
                                    each node in the agent network the service
                                    wants the client ot know about.
        """
        self._log_agent_network_usage("connectivity")
        _ = request_dict
        response_dict: Dict[str, Any] = {
        }

        reporter = ConnectivityReporter(self.agent_network)
        config: Dict[str, Any] = self.agent_network.get_config()
        metadata: Dict[str, Any] = config.get("metadata")
        connectivity_info: List[Dict[str, Any]] = reporter.report_network_connectivity()
        response_dict = {
            "connectivity_info": connectivity_info,
        }
        if metadata is not None:
            response_dict["metadata"] = metadata

        return response_dict

    # pylint: disable=too-many-locals,too-many-statements,too-many-branches
    async def streaming_chat(self, request_dict: Dict[str, Any]) -> AsyncGenerator[Dict[str, Any], None]:
        """
        :param request_dict: A dictionary version of the ChatRequest
                    protobufs structure. Has the following keys:
            "user_message" - A ChatMessage dict representing the user input to the chat stream
            "chat_context" - A ChatContext dict representing the state of the previous conversation
                            (if any)
        :return: An iterator of dictionary versions of the ChatResponse
                    protobufs structure. Has the following keys:
            "response"      - An optional ChatMessage dictionary.  See chat.proto for details.

            Note that responses to the chat input might be numerous and will come as they
            are produced until the system decides there are no more messages to be sent.
        """
        self._log_agent_network_usage("streaming_chat")
        extractor = DictionaryExtractor(request_dict)

        # Get the user input.
        user_input = extractor.get("user_message.text")

        # Create the gateway to the internals.
        chat_session = DataDrivenChatSession(agent_network=self.agent_network)

        # Prepare the response dictionary
        template_response_dict: Dict[str, Any] = {}

        if chat_session is None or user_input is None:
            # Can't go on to chat, so report back early with a single value.
            # There is no ChatMessage response in the dictionary in this case
            yield template_response_dict
            return

        # Create a message filter so as to minimize network traffic per what the user wants
        chat_filter: Dict[str, Any] = request_dict.get("chat_filter")
        message_filter: MessageFilter = MessageFilterFactory.create_message_filter(chat_filter)

        chat_context: Dict[str, Any] = request_dict.get("chat_context")
        sly_data: Dict[str, Any] = request_dict.get("sly_data")

        # Create an asynchronous background task to process the user input.
        # This might take a few minutes, which can be longer than some
        # sockets stay open.
        asyncio_executor: AsyncioExecutor = self.invocation_context.get_asyncio_executor()
        future: Future = asyncio_executor.submit(self.request_id, chat_session.streaming_chat,
                                                 user_input, self.invocation_context, sly_data,
                                                 chat_context)
        # Ignore the future. Live in the now.
        _ = future

        # Late-stage conversions for any and all messages
        message_processor: MessageProcessor = chat_session.create_outgoing_message_processor()

        # The generator below will asynchronously block waiting for
        # chat.ChatMessage dictionaries to come back asynchronously from the submit()
        # above until there are no more from the input.
        queue_generator = self.invocation_context.get_queue()
        try:
            # Logic of what is done here:
            # 1. We tell underlying chat_session to delete its resources since we are done with this request;
            # 2. We do this in "finally" block so this releasing of resources happens in any case,
            #    including getting our async generator
            #    (which is implicitly constructed by these code lines and returned by this method)
            #    interrupted by caller-side "aclose" method.
            # 3. And we suppress all exceptions while deleting resources to keep things quieter.
            async for message in queue_generator:
                response_dict: Dict[str, Any] = copy(template_response_dict)
                if message_filter.allow(message):
                    # We expect the message to be a dictionary form of chat.ChatMessage
                    if message_processor is not None:
                        message_type: ChatMessageType = message.get("type")
                        # Can modify message
                        await message_processor.async_process_message(message, message_type)
                    response_dict["response"] = message
                    yield response_dict
        finally:
            # Release resources without exceptions
            with suppress(Exception):
                await chat_session.delete_resources()

    def reset(self):
        """
        Allows for re-use of the same instance for clients
        """
        self.invocation_context.reset()

    def close(self):
        """
        Tears down resources created
        """
        if self.invocation_context is None:
            return
        self.invocation_context.close()
        self.invocation_context = None




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any

from neuro_san.message_processing.basic_message_processor import BasicMessageProcessor


class AgentEvaluator:
    """
    Interface definition for evaluating part of an agent's response
    """

    def evaluate(self, processor: BasicMessageProcessor, test_key: str, verify_for: Any):
        """
        Evaluate the contents of the BasicMessageProcessor

        :param processor: The BasicMessageProcessor to evaluate
        :param test_key: the compound .-delimited key of the response value to test
        :param verify_for: The data to evaluate the response against
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Iterable
from typing import Tuple

import os
import random
import time
from functools import partial

from json import dumps
from json import loads
from json.decoder import JSONDecodeError
from logging import getLogger
from logging import Logger

from boto3 import client as boto3_client
from botocore.client import BaseClient
from botocore.exceptions import BotoCoreError
from botocore.exceptions import ClientError
from botocore.exceptions import NoCredentialsError

from neuro_san.interfaces.reservation import Reservation
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.network_providers.abstract_reservations_storage \
    import AbstractReservationsStorage
from neuro_san.internals.reservations.reservation_dictionary_converter import ReservationDictionaryConverter


class S3ReservationsStorage(AbstractReservationsStorage):
    """
    AWS S3-based implementation of ReservationsStorage.

    Stores reservations as JSON objects in an S3 bucket, with each reservation
    stored in its associated agent spec as metadata.
    """
    # pylint: disable=too-many-instance-attributes

    def __init__(self, bucket_name: str = "", prefix: str = "reservations/",
                 check_expirations_interval_seconds: float = 0.0):
        """
        Initialize S3 reservations storage.

        :param bucket_name: S3 bucket name (defaults to AGENT_RESERVATIONS_S3_BUCKET env var)
        :param prefix: S3 key prefix for reservation objects
        :param check_expirations_interval_seconds: How often to check for expired reservations.
                                    If 0 or negative, expiration checks are disabled.
        """
        # Our default for check_expirations_interval_seconds is 0
        # because S3 expiration check is generally a significant execution load,
        # and we may want to run it externally on demand rather than on a fixed schedule inside the service.
        super().__init__(storage_name="s3_storage",
                         check_expirations_interval_seconds=check_expirations_interval_seconds)
        self.logger: Logger = getLogger(self.__class__.__name__)

        # Check if expiration interval is set by environment variable,
        # and adjust it if so (overriding the constructor parameter)
        envvar_name: str = "AGENT_RESERVATIONS_EXTERNAL_STORAGE_CHECK_PERIOD_SECONDS"
        envvar_value: str = os.getenv(envvar_name, "0")
        try:
            expiration_check_period_seconds: float = float(envvar_value)
            self._check_interval_seconds = expiration_check_period_seconds
        except ValueError as exc:
            self.logger.error(
                "Invalid value for %s, must be a number. Got: %s. "
                "Please correct the environment variable or unset it.",
                envvar_name,
                envvar_value,
            )
            raise ValueError(
                f"Invalid value for {envvar_name}: expected a numeric value, got {envvar_value!r}"
            ) from exc

        # Configure bucket name from parameter or environment variable
        env_bucket: str = os.getenv("AGENT_RESERVATIONS_S3_BUCKET", "")
        self.bucket_name: str = bucket_name or env_bucket
        if not self.bucket_name:
            raise ValueError(
                "S3 bucket name must be provided via bucket_name parameter or "
                "AGENT_RESERVATIONS_S3_BUCKET environment variable"
            )

        # Set up S3 key prefix and initialize sync target
        self.prefix: str = prefix
        self.s3_client: BaseClient = None

        # Track last sync timestamp for incremental syncing (0.0 means sync all)
        self.last_sync_timestamp: float = 0.0
        self.converter = ReservationDictionaryConverter()
        self.max_keys_per_page: int = 1000  # Max allowed by S3 API for ListObjectsV2

    def start(self):
        """
        Initialize the S3 client and validate connection to the bucket.

        This method can be called to re-initialize the connection if needed.
        """
        try:
            # Initialize S3 client using default AWS credential chain
            self.s3_client = boto3_client("s3")

            # Validate bucket exists and we have access by performing a head operation
            self.s3_client.head_bucket(Bucket=self.bucket_name)
            self.logger.info("%s: Successfully connected to S3 bucket: %s", self._name, self.bucket_name)

        except NoCredentialsError as exception:
            # Handle missing AWS credentials
            raise ValueError(f"{self._name}: AWS credentials not found. Please configure AWS credentials.") \
                from exception
        except ClientError as exception:
            # Handle various S3 access errors with specific messages
            error_code: str = exception.response["Error"]["Code"]
            if error_code == "404":
                raise ValueError(f"{self._name}: S3 bucket '{self.bucket_name}' does not exist") from exception
            if error_code == "403":
                raise ValueError(f"{self._name}: Access denied to S3 bucket '{self.bucket_name}'") from exception
            raise ValueError(f"{self._name}: Error accessing S3 bucket '{self.bucket_name}': {exception}") \
                from exception
        # We are good with S3 connection and bucket access at this point,
        # let's start underlying logic:
        super().start()

    def _is_retryable_client_error(self, err: ClientError) -> bool:
        """
        Determine if a ClientError is worth retrying based on its error code and HTTP status.
        :param err: The ClientError exception to evaluate
        :return: True if the error is likely transient and worth retrying, False otherwise
        """
        client_error = err.response.get("Error")
        if not client_error:
            client_error = {}
        code = client_error.get("Code", "")
        error_response = err.response.get("ResponseMetadata")
        if not error_response:
            error_response = {}
        status = error_response.get("HTTPStatusCode", 0)

        # Common codes for transient situations
        retryable_codes = {
            "SlowDown",
            "Throttling",
            "ThrottlingException",
            "RequestTimeout",
            "RequestTimeoutException",
            "InternalError",
            "ServiceUnavailable",
            "503",
        }
        if code in retryable_codes:
            return True
        # Retry on some 5xx
        if isinstance(status, int) and 500 <= status < 600:
            return True
        return False

    def _do_with_retries(self, fn, *, max_attempts: int = 8, base_sleep: float = 0.25):
        """
        Generic retry wrapper for boto3 calls.
        boto3/botocore already retries, but this adds a bit of extra resilience and backoff for batch operations.

        """
        attempt: int = 1
        while True:
            try:
                return fn()
            except ClientError as err:
                if attempt >= max_attempts or not self._is_retryable_client_error(err):
                    raise
                # Compute exponential backoff with jitter
                sleep = base_sleep * (2 ** (attempt - 1))
                sleep = sleep * (0.5 + random.random())  # sleep time jitter
                self.logger.warning("%s: Retryable ClientError (%s). attempt=%d", self._name, err, attempt)
                time.sleep(sleep)
                attempt += 1
            except BotoCoreError as err:
                # Often transient network/serialization issues
                if attempt >= max_attempts:
                    raise
                # Compute exponential backoff with jitter
                sleep = base_sleep * (2 ** (attempt - 1))
                sleep = sleep * (0.5 + random.random())
                self.logger.warning("%s: Retryable BotoCoreError (%s). attempt=%d", self._name, err, attempt)
                time.sleep(sleep)
                attempt += 1

    def get_obj_key_for_reservation(self, reservation_id: str) -> str:
        """
        Helper method to construct the S3 object key for a given reservation ID.
        :param reservation_id: The ID of the reservation
        :return: The corresponding S3 object key
        """
        return f"{self.prefix}{reservation_id}.json"

    def add_reservations(self, reservations_dict: Dict[Reservation, Any],
                         source: str = None):
        """
        Add reservations to S3 storage.

        :param reservations_dict: A mapping of Reservation -> some deployable agent spec
        :param source: A string describing where the deployment was coming from
        """
        self.logger.info("%s: Adding %d reservations to S3", self._name, len(reservations_dict))

        # Process each reservation/agent spec pair individually
        reservation: Reservation = None
        agent_spec: Dict[str, Any] = None
        for reservation, agent_spec in reservations_dict.items():

            # Build complete data structure containing reservation metadata,
            # the associated agent_spec, source information, and storage timestamp
            current_time: float = time.time()
            new_metadata: Dict[str, Any] = {
                "reservation": self.converter.to_dict(reservation),  # Serialized reservation object
                "stored_at": current_time              # When stored in S3
            }
            if agent_spec.get("metadata") is None:
                agent_spec["metadata"] = {}
            agent_spec["metadata"].update(new_metadata)

            # Generate S3 key using prefix and reservation ID for easy lookup
            reservation_id: str = reservation.get_reservation_id()
            key: str = self.get_obj_key_for_reservation(reservation_id)

            # Store as JSON object in S3 with proper content type
            json_body: str = dumps(agent_spec, indent=4)  # Pretty-printed JSON

            put_function = partial(self.s3_client.put_object,
                                   Bucket=self.bucket_name,
                                   Key=key,
                                   Body=json_body,
                                   ContentType="application/json")
            self._do_with_retries(put_function)

            self.logger.debug("%s: Successfully stored reservation %s in S3", self._name, reservation_id)

    def _retrieve_object_with_retries(self, obj_key: str) -> Dict[str, Any]:
        """
        Helper method to retrieve an S3 object with retries.
        :param obj_key: S3 object key to retrieve
        :return: The parsed JSON content of the S3 object as a dictionary
        :raises: ClientError if the object cannot be retrieved after retries
                 JSONDecodeError if the content cannot be parsed as JSON
        """
        get_function = partial(self.s3_client.get_object, Bucket=self.bucket_name, Key=obj_key)
        obj_response: Dict[str, Any] = self._do_with_retries(get_function)
        # Parse JSON content from S3 object body
        json_content: str = obj_response["Body"].read().decode("utf-8")
        return loads(json_content)

    def get_one_reservation(self, obj_key: str) -> Tuple[Reservation, Any]:
        """
        Sync a single reservation from S3.

        :param obj_key: reservation ID to retrieve (used to construct S3 object key)
        :return: Tuple of (reservation, agent_spec) if successful and not expired,
                 (None, None) otherwise
        """
        reservation: Reservation = None
        agent_network: AgentNetwork = None
        # Construct the S3 object key for this reservation ID
        s3_obj_key: str = self.get_obj_key_for_reservation(obj_key)
        try:
            # Retrieve the reservation object from S3
            agent_spec: Dict[str, Any] = self._retrieve_object_with_retries(s3_obj_key)
            metadata: Dict[str, Any] = agent_spec.get("metadata")

            # Reconstruct the Reservation object from stored dictionary
            reservation_dict: Dict[str, Any] = metadata.get("reservation")
            reservation = self.converter.from_dict(reservation_dict)
            # Reconstruct the AgentNetwork object using the agent spec dictionary
            # and reservation ID - which is our agent name in this design
            agent_network: AgentNetwork = AgentNetwork(agent_spec, reservation.get_reservation_id())

            self.logger.debug("%s: Successfully synced active reservation %s",
                              self._name, reservation.get_reservation_id())

        except ClientError as exception:
            # Handle case where another process already removed the object before we could read it
            if exception.response["Error"]["Code"] == "NoSuchKey":
                self.logger.debug("%s: Reservation %s was already removed by another process during sync",
                                  self._name, obj_key)
            else:
                # Log other S3 errors but don't raise - allows sync to continue
                self.logger.error("%s: S3 error processing reservation object %s during sync: %s",
                                  self._name, obj_key, str(exception))

        except JSONDecodeError as exception:
            # Log JSON errors but don't raise - allows sync to continue
            self.logger.error("%s: JSON error processing reservation object %s during sync: %s",
                              self._name, obj_key, str(exception))

        return reservation, agent_network

    def expire_one_reservation(self, obj_key: str, current_time: float) -> bool:
        """
        Check and expire a single reservation if it's expired.

        :param obj_key: S3 object key for the reservation
        :param current_time: Current timestamp to compare against
        :return: True if reservation was expired and deleted, False otherwise
        """
        expired: bool = False
        try:
            # Retrieve the reservation object from S3
            agent_spec: Dict[str, Any] = self._retrieve_object_with_retries(obj_key)
            metadata: Dict[str, Any] = agent_spec.get("metadata")
            reservation_data: Dict[str, Any] = metadata.get("reservation")

            # Compare current time against reservation's expiration timestamp
            expiration_time: float = reservation_data.get("expiration_time_in_seconds")
            if current_time > expiration_time:
                # Reservation has expired - remove it from S3 storage
                try:
                    delete_function = partial(self.s3_client.delete_object, Bucket=self.bucket_name, Key=obj_key)
                    self._do_with_retries(delete_function)
                    reservation_id: str = reservation_data.get("id")
                    self.logger.debug("%s: Deleted expired reservation %s from S3", self._name, reservation_id)
                    expired = True
                except ClientError as delete_error:
                    # Handle case where another process already deleted the object
                    if delete_error.response["Error"]["Code"] != "NoSuchKey":
                        # Re-raise other delete errors
                        raise delete_error

                    self.logger.debug("%s: Reservation %s was already deleted by another process", self._name, obj_key)
                    expired = True  # Consider this a successful expiration

            # Reservation is still active - no action needed

        except ClientError as exception:
            # Handle case where another process already removed the object before we could read it
            if exception.response["Error"]["Code"] == "NoSuchKey":
                self.logger.debug("%s: Reservation %s was already removed by another process", self._name, obj_key)
                expired = True  # Object is gone, which is the desired outcome for expiration
            else:
                # Log other S3 errors but don't raise - allows expiration to continue
                self.logger.error("%s: S3 error processing reservation object %s: %s",
                                  self._name, obj_key, str(exception))

        except JSONDecodeError as exception:
            # Log JSON errors but don't raise - allows expiration process to continue
            self.logger.error("%s: JSON error processing reservation object %s during expire: %s",
                              self._name, obj_key, str(exception))

        return expired

    def iter_reservation_keys(self) -> Iterable[str]:
        """
        Lists ALL objects (under current S3 bucket prefix) using a paginator.
        Yields object keys.
        """
        paginator = self.s3_client.get_paginator("list_objects_v2")
        paginate_function = partial(paginator.paginate,
                                    Bucket=self.bucket_name,
                                    Prefix=self.prefix,
                                    PaginationConfig={"PageSize": self.max_keys_per_page})
        for page in self._do_with_retries(paginate_function):
            # _do_with_retries here wraps the generator creation
            contents = page.get("Contents", [])
            for obj in contents:
                yield obj["Key"]

    def expire_reservations(self):
        """
        Remove expired reservations from S3 storage.
        """
        self.logger.debug("%s: Starting expiration process for S3 reservations", self._name)

        # Track how many reservations we expire for reporting
        expired_count: int = 0
        # Get current timestamp once for consistent expiration checking
        current_time: float = time.time()

        for reservation_key in self.iter_reservation_keys():
            # Attempt to expire this reservation and increment counter if successful
            if self.expire_one_reservation(reservation_key, current_time):
                expired_count += 1

        if expired_count > 0:
            self.logger.info("%s: Expiration complete: removed %d expired reservations from S3",
                             self._name, expired_count)
        else:
            self.logger.debug("%s: Expiration complete: removed no expired reservations from S3", self._name)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

from logging import getLogger
from logging import Logger

from langchain.agents.middleware.types import AgentMiddleware
from langchain_core.messages import BaseMessage

from leaf_common.config.resolver_util import ResolverUtil
from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san.interfaces.reservationist import Reservationist
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.originating_journal import OriginatingJournal
from neuro_san.internals.journals.progress_journal import ProgressJournal
from neuro_san.internals.messages.origination import Origination
from neuro_san.internals.reservations.accumulating_agent_reservationist import AccumulatingAgentReservationist

# We don't want to drag in langgraph as a dependency, but we will allow Checkpointers from there.
Checkpointer: Type = Any


class MiddlewareFactory:
    """
    Creates instances of AgentMiddleware based on a list of middleware configs.
    """

    def __init__(self, invocation_context: InvocationContext,
                 origin: List[Dict[str, Any]],
                 chat_history: List[BaseMessage]):
        """
        :param invocation_context: The invocation context for the request.
        :param origin: A list of dictionaries containing origin information as to which agent
                    is creating the middleware.
        :param chat_history: The chat history of the RunContext that the middleware is being created for.
        """
        self.invocation_context: InvocationContext = invocation_context
        self.origin: List[Dict[str, Any]] = origin
        self.chat_history: List[BaseMessage] = chat_history
        self.logger: Logger = getLogger(self.__class__.__name__)

    def create_agent_middleware(self, config: List[Dict[str, Any]], sly_data: Dict[str, Any]) \
            -> Tuple[List[AgentMiddleware], Checkpointer]:
        """
        Create instances of AgentMiddleware based on the config.
        :param config: A list of dictionaries containing specific middleware configurations.
                    Order matters.
        :param sly_data: The private sly_data dictionary for the agent invocation
        :return: A tuple of:
                * A list of instances of AgentMiddleware.
                  Can be an empty list if there is no middleware to create.
                * A Checkpointer instance if required by the config. Can be None.
        """
        # Keep a running list of what we've created.
        checkpointer: Checkpointer = None
        all_middleware: List[AgentMiddleware] = []

        if config is None or len(config) == 0:
            # No middleware specified
            return all_middleware, checkpointer

        # Loop through the config entries.
        for middleware_index, one_config in enumerate(config):

            middleware: AgentMiddleware = None
            created_checkpointer: Checkpointer = None
            middleware, created_checkpointer = self.create_middleware(one_config, sly_data, middleware_index)
            if middleware is not None:
                all_middleware.append(middleware)

            if created_checkpointer is not None:
                if checkpointer is not None:
                    self.logger.warning("Multiple Checkpointers specified. Using the first one.")
                else:
                    checkpointer = created_checkpointer

        return all_middleware, checkpointer

    def create_middleware(self, config: Dict[str, Any], sly_data: Dict[str, Any], index: int = 0) \
            -> Tuple[AgentMiddleware, Checkpointer]:
        """
        Create an instance of AgentMiddleware based on the config.

        :param config: A dictionary containing specific middleware configurations.
        :param sly_data: The private sly_data dictionary for the agent invocation.
        :param index: The index of the middleware in the config
        :return:
                Can be None if there is no middleware to create.
        :return: A tuple of:
                * An instance of AgentMiddleware per the config.
                  Can be None if there is no middleware to create.
                * A Checkpointer instance if required by the config. Can be None.
        """
        checkpointer: Checkpointer = None
        middleware: AgentMiddleware = self.create_class("AgentMiddleware", AgentMiddleware, config, sly_data, index)

        # If needed, we could do some kind of kwargs dict here instead of a specific checkpointer arg for create_agent()
        checkpointer_config = config.get("checkpointer")
        if checkpointer_config is not None:
            checkpointer = self.create_class("Checkpointer", Checkpointer, checkpointer_config, sly_data, index)

        return middleware, checkpointer

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def create_class(self, what_we_are_creating: str, create_type: Type[Any],
                     config: Dict[str, Any], sly_data: Dict[str, Any], index: int) -> Any:
        """
        Create an instance of a class based on the config.

        :param what_we_are_creating: A string to describe what we are creating.
        :param create_type: The type of class to create.
        :param config: A dictionary containing specific class configurations.
        :param sly_data: The private sly_data dictionary for the agent invocation.
        :param index: The index of the class in the config.
        :return: An instance of the class or None if class creation fails.
        """
        error_context: str = self._build_error_context(what_we_are_creating, index)

        class_name: str = self._get_class_name(config, error_context)
        if class_name is None:
            return None

        created_class: Type[Any] = self._resolve_class(class_name, error_context, create_type)

        middleware_origin, middleware_origin_str, reservationist = self._prepare_middleware_context(class_name, config)

        args: Dict[str, Any] = self._prepare_args(
            config, sly_data, middleware_origin, middleware_origin_str, reservationist)

        return self._instantiate_class(created_class, args, what_we_are_creating, class_name)

    def _build_error_context(self, what_we_are_creating: str, index: int) -> str:
        """Build a descriptive error context string."""
        origin_str: str = Origination.get_full_name_from_origin(self.origin)
        return f"{what_we_are_creating} at index {index} of {origin_str}"

    def _get_class_name(self, config: Dict[str, Any], error_context: str) -> Optional[str]:
        """
        Validate config and extract the class name.
        :return: The class name, or None if config is invalid.
        """
        if not isinstance(config, dict):
            self.logger.warning("%s is missing a configuration dictionary. Skipping it.", error_context)
            return None

        class_name: str = config.get("class")
        if class_name is None:
            self.logger.warning("%s is missing the 'class' field. Skipping it.", error_context)
            return None

        return class_name

    def _resolve_class(self, class_name: str, error_context: str, create_type: Type[Any]) -> Type[Any]:
        """Resolve and return the class from the class name."""
        try:
            return ResolverUtil.create_class(class_name, error_context, create_type)
        except ValueError as exception:
            message = f"""
Problems loading class {class_name}: {str(exception)}

Check these things:
1. Is the class findable from what is set in your PYTHONPATH?
"""
            self.logger.error(message)
            raise ValueError(message) from exception

    def _prepare_middleware_context(self, class_name: str, config: Dict[str, Any]) \
            -> Tuple[List[Dict[str, Any]], str, Reservationist]:
        """
        Prepare the origin and reservationist for the middleware.
        :param class_name: The class name of the middleware.
        :param config: The middleware configuration dictionary.
        :return: A tuple of (middleware_origin, middleware_origin_str, reservationist).
        """
        tool_name_for_class: str = class_name.split(".")[-1]
        origination: Origination = self.invocation_context.get_origination()
        middleware_origin: List[Dict[str, Any]] = origination.add_spec_name_to_origin(self.origin, tool_name_for_class)
        middleware_origin_str: str = Origination.get_full_name_from_origin(middleware_origin)

        reservationist: Reservationist = None
        config_extractor = DictionaryExtractor(config)
        if config_extractor.get("allow.reservations"):
            real_reservationist: Reservationist = self.invocation_context.get_reservationist()
            if real_reservationist is not None:
                reservationist = AccumulatingAgentReservationist(real_reservationist, middleware_origin_str)

        return middleware_origin, middleware_origin_str, reservationist

    def _prepare_args(self, config: Dict[str, Any], sly_data: Dict[str, Any],
                      middleware_origin: List[Dict[str, Any]],
                      middleware_origin_str: str, reservationist: Reservationist) -> Dict[str, Any]:
        """
        Prepare the args dict, injecting framework-provided values for special keys.
        """
        args: Dict[str, Any] = dict(config.get("args", {}))

        # Map of special arg keys to their framework-provided values.
        special_args: Dict[str, Any] = {
            "origin": middleware_origin,
            "origin_str": middleware_origin_str,
            "reservationist": reservationist,
            "sly_data": sly_data,
            "chat_history": self.chat_history,
        }
        for key, value in special_args.items():
            if key in args:
                args[key] = value

        # Journal-based args require lazy creation of the journal.
        if "journal" in args or "progress_reporter" in args:
            journal = OriginatingJournal(self.invocation_context.get_journal(), middleware_origin)
            if "journal" in args:
                args["journal"] = journal
            if "progress_reporter" in args:
                args["progress_reporter"] = ProgressJournal(journal)

        return args

    def _instantiate_class(self, created_class: Type[Any], args: Dict[str, Any],
                           what_we_are_creating: str, class_name: str) -> AgentMiddleware:
        """Instantiate the resolved class with the prepared args."""
        try:
            return created_class(**args)
        except TypeError as exception:
            message = f"""
Problems creating {what_we_are_creating} instance of {class_name}: {str(exception)}

Check these things:
1. Is the class findable from what is set in your PYTHONPATH?
2. Are the args correct?
"""
            self.logger.error(message)
            raise ValueError(message) from exception




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import List

from neuro_san.internals.interfaces.dictionary_validator import DictionaryValidator
from neuro_san.internals.validation.common.composite_dictionary_validator import CompositeDictionaryValidator
from neuro_san.internals.validation.network.cycles_network_validator import CyclesNetworkValidator
from neuro_san.internals.validation.network.missing_nodes_network_validator import MissingNodesNetworkValidator
from neuro_san.internals.validation.network.unreachable_nodes_network_validator import UnreachableNodesNetworkValidator


class StructureNetworkValidator(CompositeDictionaryValidator):
    """
    Implementation of CompositeDictionaryValidator interface that uses multiple specific validators
    to do some standard validation for topological issues.
    This gets used by agent network designer.
    """

    def __init__(self):
        """
        Constructor
        """
        validators: List[DictionaryValidator] = [
            CyclesNetworkValidator(),
            MissingNodesNetworkValidator(),
            UnreachableNodesNetworkValidator(),
        ]
        super().__init__(validators)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from pathlib import Path, PurePosixPath

from neuro_san.internals.interfaces.agent_name_mapper import AgentNameMapper


class AgentFileTreeMapper(AgentNameMapper):
    """
    A simple policy implementation defining conversion
    between agent name as specified in a manifest file
    and a file path (relative to registry root directory) to this agent definition file.
    """
    def __init__(self, path_method=Path):
        """
        Constructor

        :param path_method: Optional Path method to use for path manipulations.
            Default is pathlib.Path, but can be overridden for testing purposes.
        """
        self.path_method = path_method

    def agent_name_to_filepath(self, agent_name: str) -> str:
        """
        Converts an agent name from registry manifest to file path to this agent definition file.
        """
        # agent_name from registry manifest uses '/' as separator,
        # we need to convert it to the local OS file path format.
        return str(self.path_method(PurePosixPath(agent_name)))

    def filepath_to_agent_network_name(self, filepath: str) -> str:
        """
        Converts a file path to agent definition file (relative to registry root directory)
        to agent network name identifying it to the service.
        """
        # Remove file name extension, convert it to "/" separator.
        # Specifically:
        # self.path_method(filepath).as_posix() - converts to Path format we use (default is OS native)
        #     and then to posix format (with '/' separator). This gets us a string.
        # PurePosixPath(...).with_suffix("") - converts string to a posix path (with "/")
        #     and removes the file extension.
        # str(...) - converts back to string.
        return str(PurePosixPath(self.path_method(filepath).as_posix()).with_suffix(""))




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import Type

from pydantic import BaseModel
from pydantic.v1 import Field
from pydantic.v1 import create_model

from leaf_common.serialization.interface.dictionary_converter import DictionaryConverter


class BaseModelDictionaryConverter(DictionaryConverter):
    """
    DictionaryConverter implementation which can convert an OpenAI
    function spec (the dictionary) into a pydantic BaseModel
    which describes the same object structure or "shape".
    """

    TYPE_LOOKUP: Dict[str, Type] = {
        "string": str,
        "int": int,
        "float": float,
        "boolean": bool,
        "array": List,

        # Note: "Any" produces a pydantic BaseModel object whose fields
        #       are direct members when passed as an argument to a CodedTool
        #       or any other Tool.  If you really want a dictionary,
        #       you need to turn it into one using foo = dict(my_object_arg)
        #       See: https://docs.pydantic.dev/1.10/usage/exporting_models/#dictmodel-and-iteration
        #       This is what the PydanticArgumentDictionaryConverter is for.
        "object": Any
    }

    def __init__(self, top_level_field_name: str):
        """
        Constructor

        :param top_level_field_name: The field name for the top-level object
        """
        self.top_level_field_name: str = top_level_field_name

    def to_dict(self, obj: BaseModel) -> Dict[str, Any]:
        """
        :param obj: The object to be converted into a dictionary
        :return: A data-only dictionary that represents all the data for
                the given object, either in primitives
                (booleans, ints, floats, strings), arrays, or dictionaries.
                If obj is None, then the returned dictionary should also be
                None.  If obj is not the correct type, it is also reasonable
                to return None.
        """
        # At this point we are not going back to OpenAI functional specs
        raise NotImplementedError

    def from_dict(self, obj_dict: Dict[str, Any]) -> BaseModel:
        """
        :param obj_dict: The data-only dictionary to be converted into an object
        :return: An object instance created from the given dictionary.
                If obj_dict is None, the returned object should also be None.
                If obj_dict is not the correct type, it is also reasonable
                to return None.
        """
        base_model: BaseModel = self.openai_function_to_pydantic(self.top_level_field_name, obj_dict)
        return base_model

    def openai_function_to_pydantic(self, name: str, function_dict: Dict[str, Any]) -> BaseModel:
        """
        Turns an openai function spec dictionary into a pydantic BaseModel
        See: https://docs.pydantic.dev/latest/concepts/models/#dynamic-model-creation

        :param name: The string name of the object/field undergoing conversion
        :param function_dict: The dictionary describing the OpenAI function to
                    be converted into a pydantic BaseModel
        :return: The pydantic BaseModel that corresponds to the OpenAI function spec.
        """

        # Get stuff from the object-level function dictionary
        # from the OpenAI function spec
        properties: Dict[str, Any] = function_dict.get("properties")
        required: List[str] = []
        required = function_dict.get("required", required)

        fields: Dict[str, Any] = {}
        for field_name, one_property in properties.items():

            # Get bits we need to assemble a pydantic Field description
            description: str = one_property.get("description")
            field_type: Type = self.get_type_from_property_dict(field_name, one_property)

            # Assemble a pydantic Field
            field_kwargs: Dict[str, Any] = {}

            # Description helps the agents communicate upstream what the field does
            if description is not None:
                field_kwargs["description"] = description

            # Setting a default to None is like saying that it's not required.
            # This allows agent infra to not include the field in the args
            # when a value is not there.  By contrast, we do not set a default
            # for something that is required, which leaves the pydantic definition
            # of "Undefined" in place which to agent infra implies a required arg.
            if field_name not in required:
                field_kwargs["default"] = None
            field = Field(**field_kwargs)

            # Add the field to our dictionary with its name as key
            fields[field_name] = (field_type, field)

        # Create the pydantic BaseModel for the type dynamically
        model: BaseModel = create_model(name, **fields)
        return model

    def get_type_from_property_dict(self, field_name: str, one_property: Dict[str, Any]) -> Type:
        """
        :param field_name: The string name of the field undergoing conversion
        :param one_property: The property dictionary of the field whose type we are looking for.
        :return: the Type of the property to be used with pydantic Fields
                This type may be a BaseModel if object specification is
                deep enough.
        """
        type_from_dict: str = one_property.get("type")
        field_type: Type = self.TYPE_LOOKUP.get(type_from_dict)

        if type_from_dict == "object":

            object_props: Dict[str, Any] = one_property.get("properties")
            if object_props is not None and object_props:
                # If the object has properties, make a new pydantic BaseModel for it
                # and include that as its type. This allows field descriptions,
                # and required-ness to be part of the object definition.
                field_type = self.openai_function_to_pydantic(field_name, one_property)

            # If there are no properties, we use the default in the TYPE_LOOKUP
            # which is Any.  Any passes a dynamically constructed pydantic object
            # whose fields can be accessed like regular fields.  This is nice
            # and fancy, but what we really want is a dictionary.

        elif type_from_dict == "array":

            # Get the type of the list components/items
            items: Dict[str, Any] = one_property.get("items")
            item_type: Type = self.get_type_from_property_dict(f"{field_name}_component", items)

            # Set the field_type as a properly typed generic List
            field_type = List[item_type]

        return field_type
